@@ -1,4 +1,4 @@
-// Copyright (C) 2014-2017 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2014-2019 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -325,7 +325,11 @@ public:
     /// lease has been later updated (e.g. as a result of Rebind) as it is
     /// expected that the fresh lease has cltt set to "now" (not to the time
     /// in the past).
-    void fastFwdTime(const uint32_t secs);
+    ///
+    /// @param secs Number of seconds by which the time should be moved.
+    /// @param update_server Indicates if the leases should be updated on the
+    /// server.
+    void fastFwdTime(const uint32_t secs, const bool update_server = false);
 
     /// @brief Returns DUID option used by the client.
     OptionPtr getClientId() const;
@@ -495,14 +499,29 @@ public:
 
     /// @brief Returns status code set by the server for the IAID.
     ///
-    /// @warning This method doesn't check if the specified index is out of
-    /// range. The caller is responsible for using a correct offset by
-    /// invoking the @c getLeaseNum function.
-    ///
-    /// @param at Index of the lease held by the client.
-    /// @return A status code for the lease at the specified index.
+    /// @param iaid for which the status is desired
+    /// @return A status code for the given iaid
     uint16_t getStatusCode(const uint32_t iaid) const;
 
+    /// @brief Returns T1 and T2 timers associated with a given iaid
+    ///
+    /// Currently this method gleans the T1 an T2 times from the first
+    /// aquired lease belonging to the target iaid.  Since all leases for
+    /// an iaid should have the same values for T1 and T2, this should be 
+    /// fine.  If there are no acquired leases for the iaid, then t1 and
+    /// t2 are indeterminate.
+    ///
+    /// The primary impetus for this method is isolate the fact that T1
+    /// and T2 are stored on each lease.  This may change in the future. 
+    ///
+    /// @param iaid  iaid of the target IA
+    /// @param[out] t1 set to the value of the IA's T1
+    /// @param[out] t2 set to the value of the IA's T2
+    /// if there are no leases for the iaid
+    ///
+    /// @return true if there are aquired leases for the given iaid
+    bool getTeeTimes(const uint32_t iaid, uint32_t& t1, uint32_t& t2) const;
+
     /// @brief Returns number of acquired leases.
     size_t getLeaseNum() const {
         return (config_.leases_.size());
@@ -582,7 +601,7 @@ public:
     ///
     /// This method is typically called to specify IA_NA options to be
     /// sent to the server during 4-way handshakes and during lease
-    /// renewal to request allocation of new leases (as per RFC7550).
+    /// renewal to request allocation of new leases (as per RFC 8415).
     ///
     /// @param iaid IAID.
     /// @param address IPv6 address to be included in the IA_NA. It defaults
@@ -603,7 +622,7 @@ public:
     ///
     /// This method is typically called to specify IA_PD options to be
     /// sent to the server during 4-way handshakes and during lease
-    /// renewal to request allocation of new leases (as per RFC7550).
+    /// renewal to request allocation of new leases (as per RFC 8415).
     ///
     /// @param iaid IAID.
     /// @param prefix_len Prefix length.
@@ -639,6 +658,11 @@ public:
         relay_link_addr_ = link_addr;
     }
 
+    /// @brief Sets interface id value to be inserted into relay agent option.
+    ///
+    /// @param interface_id Value of the interface id as string.
+    void useInterfaceId(const std::string& interface_id);
+
     /// @brief Controls whether the client should send a client-id or not
     /// @param send should the client-id be sent?
     void useClientId(const bool send) {
@@ -741,6 +765,32 @@ public:
     /// @param opt additional option to be sent
     void addExtraOption(const OptionPtr& opt);
 
+    /// @brief Configures the client to not send any extra options.
+    void clearExtraOptions();
+
+    /// @brief Add a client class.
+    ///
+    /// @param client_class name of the class to be added.
+    void addClass(const ClientClass& client_class);
+
+    /// @brief Configures the client to not add client classes.
+    void clearClasses();
+
+    /// @brief Debugging method the prints currently held configuration
+    ///
+    /// @todo: This is mostly useful when debugging tests. This method
+    /// is incomplete. Please extend it when needed.
+    void printConfiguration() const;
+
+    /// @brief Receives a response from the server.
+    ///
+    /// This method is useful to receive response from the server after
+    /// parking a packet. In this case, the packet is not received as a
+    /// result of initial exchange, e.g. @c doRequest. The test can call
+    /// this method to complete the transaction when it expects that the
+    /// packet has been unparked.
+    void receiveResponse();
+
 private:
 
     /// @brief Applies the new leases for the client.
@@ -906,6 +956,12 @@ private:
 
     /// @brief FQDN requested by the client.
     Option6ClientFqdnPtr fqdn_;
+
+    /// @brief Interface id.
+    OptionPtr interface_id_;
+
+    /// @brief Extra classes to add to the query.
+    ClientClasses classes_;
 };
 
 } // end of namespace isc::dhcp::test
