@@ -1,4 +1,4 @@
-// Copyright (C) 2013-2017 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2013-2019 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -96,9 +96,23 @@ Dhcpv6SrvTest::checkIA_NA(const Pkt6Ptr& rsp, uint32_t expected_iaid,
         return (boost::shared_ptr<Option6IAAddr>());
     }
 
-    EXPECT_EQ(expected_iaid, ia->getIAID());
-    EXPECT_EQ(expected_t1, ia->getT1());
-    EXPECT_EQ(expected_t2, ia->getT2());
+    if (expected_iaid != ia->getIAID()) {
+        ADD_FAILURE() << "ia->iaid: " << ia->getIAID()
+                      << " is not expected value: " << expected_iaid;
+        return (boost::shared_ptr<Option6IAAddr>());
+    }
+
+    if (expected_t1 != ia->getT1()) {
+        ADD_FAILURE() << "ia->t1: " << ia->getT1()
+                      << " is not expected value: " << expected_t1;
+        return (boost::shared_ptr<Option6IAAddr>());
+    }
+
+    if (expected_t2 != ia->getT2()) {
+        ADD_FAILURE() << "ia->t2: " << ia->getT2()
+                      << " is not expected value: " << expected_t2;
+        return (boost::shared_ptr<Option6IAAddr>());
+    }
 
     tmp = ia->getOption(D6O_IAADDR);
     boost::shared_ptr<Option6IAAddr> addr = boost::dynamic_pointer_cast<Option6IAAddr>(tmp);
@@ -501,7 +515,7 @@ Dhcpv6SrvTest::testReleaseBasic(Lease::Type type, const IOAddress& existing,
     checkIA_NAStatusCode(ia, STATUS_Success, 0, 0);
     checkMsgStatusCode(reply, STATUS_Success);
 
-    // There should be no address returned in RELEASE (see RFC3315, 18.2.6)
+    // There should be no address returned in RELEASE (see RFC 8415, 18.3.7)
     // There should be no prefix
     EXPECT_FALSE(tmp->getOption(D6O_IAADDR));
     EXPECT_FALSE(tmp->getOption(D6O_IAPREFIX));
@@ -699,7 +713,13 @@ Dhcpv6SrvTest::configure(const std::string& config) {
 void
 Dhcpv6SrvTest::configure(const std::string& config, NakedDhcpv6Srv& srv) {
     ConstElementPtr json;
-    ASSERT_NO_THROW(json = parseJSON(config));
+    try {
+        json = parseJSON(config);
+    } catch (const std::exception& ex) {
+        // Fatal failure on parsing error
+        FAIL() << "config parsing failed, test is broken: " << ex.what();
+    }
+
     ConstElementPtr status;
 
     // Disable the re-detect flag
@@ -710,7 +730,8 @@ Dhcpv6SrvTest::configure(const std::string& config, NakedDhcpv6Srv& srv) {
     ASSERT_TRUE(status);
     int rcode;
     ConstElementPtr comment = isc::config::parseAnswer(rcode, status);
-    ASSERT_EQ(0, rcode);
+    ASSERT_EQ(0, rcode) << "configuration failed, test is broken: "
+        << comment->str();
 
     CfgMgr::instance().commit();
 }
