@@ -1,4 +1,4 @@
-// Copyright (C) 2012-2017 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2012-2019 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -14,16 +14,24 @@
 #include <dhcp6/ctrl_dhcp6_srv.h>
 #include <dhcp6/parser_context.h>
 #include <dhcp6/tests/dhcp6_test_utils.h>
+#include <dhcpsrv/cb_ctl_dhcp4.h>
 #include <dhcpsrv/cfgmgr.h>
 #include <dhcpsrv/lease.h>
 #include <dhcpsrv/lease_mgr_factory.h>
-#include <dhcpsrv/testutils/mysql_schema.h>
+#include <process/config_base.h>
+
+#ifdef HAVE_MYSQL
+#include <mysql/testutils/mysql_schema.h>
+#endif
+
 #include <log/logger_support.h>
 #include <util/stopwatch.h>
 
+#include <boost/pointer_cast.hpp>
 #include <boost/scoped_ptr.hpp>
 #include <gtest/gtest.h>
 
+#include <functional>
 #include <fstream>
 #include <iostream>
 #include <sstream>
@@ -36,16 +44,113 @@ using namespace isc;
 using namespace isc::asiolink;
 using namespace isc::config;
 using namespace isc::data;
+
+#ifdef HAVE_MYSQL
+using namespace isc::db::test;
+#endif
+
 using namespace isc::dhcp;
 using namespace isc::dhcp::test;
 using namespace isc::hooks;
 
 namespace {
 
-class NakedControlledDhcpv6Srv: public ControlledDhcpv6Srv {
-    // "Naked" DHCPv6 server, exposes internal fields
+/// @brief Test implementation of the @c CBControlDHCPv6.
+///
+/// This implementation is installed on the test server instance. It
+/// overrides the implementation of the @c databaseConfigFetch function
+/// to verify arguments passed to this function and throw an exception
+/// when desired in the negative test scenarios. It doesn't do the
+/// actual configuration fetch as this is tested elswhere and would
+/// require setting up a database configuration backend.
+class TestCBControlDHCPv6 : public CBControlDHCPv6 {
 public:
-    NakedControlledDhcpv6Srv():ControlledDhcpv6Srv(0) { }
+
+    /// @brief Constructor.
+    TestCBControlDHCPv6()
+        : CBControlDHCPv6(), db_config_fetch_calls_(0),
+          enable_check_fetch_mode_(false), enable_throw_(false) {
+    }
+
+    /// @brief Stub implementation of the "fetch" function.
+    ///
+    /// If this is not the first invocation of this function, it
+    /// verifies that the @c fetch_mode has been correctly set to
+    /// @c FetchMode::FETCH_UPDATE.
+    ///
+    /// It also throws an exception when desired by a test, to
+    /// verify that the server gracefully handles such exception.
+    ///
+    /// @param fetch_mode value indicating if the method is called upon the
+    /// server start up or it is called to fetch configuration updates.
+    ///
+    /// @throw Unexpected when configured to do so.
+    virtual void databaseConfigFetch(const process::ConfigPtr&,
+                                     const FetchMode& fetch_mode) {
+        ++db_config_fetch_calls_;
+
+        if (enable_check_fetch_mode_) {
+            if ((db_config_fetch_calls_ <= 1) && (fetch_mode == FetchMode::FETCH_UPDATE)) {
+                ADD_FAILURE() << "databaseConfigFetch was called with the value "
+                    "of fetch_mode=FetchMode::FETCH_UPDATE upon the server configuration";
+
+            } else if ((db_config_fetch_calls_ > 1) && (fetch_mode == FetchMode::FETCH_ALL)) {
+                ADD_FAILURE() << "databaseConfigFetch was called with the value "
+                    "of fetch_mode=FetchMode::FETCH_ALL during fetching the updates";
+            }
+        }
+
+        if (enable_throw_) {
+            isc_throw(Unexpected, "testing if exceptions are corectly handled");
+        }
+    }
+
+    /// @brief Returns number of invocations of the @c databaseConfigFetch.
+    size_t getDatabaseConfigFetchCalls() const {
+        return (db_config_fetch_calls_);
+    }
+
+    /// @brief Enables checking of the @c fetch_mode value.
+    void enableCheckFetchMode() {
+        enable_check_fetch_mode_ = true;
+    }
+
+    /// @brief Enables the object to throw from @c databaseConfigFetch.
+    void enableThrow() {
+        enable_throw_ = true;
+    }
+
+private:
+
+    /// @brief Counter holding number of invocations of the @c databaseConfigFetch.
+    size_t db_config_fetch_calls_;
+
+    /// @brief Boolean flag indicated if the value of the @c fetch_mode
+    /// should be verified.
+    bool enable_check_fetch_mode_;
+
+    /// @brief Boolean flag indicating if the @c databaseConfigFetch should
+    /// throw.
+    bool enable_throw_;
+};
+
+/// @brief Shared pointer to the @c TestCBControlDHCPv6.
+typedef boost::shared_ptr<TestCBControlDHCPv6> TestCBControlDHCPv6Ptr;
+
+/// @brief "Naked" DHCPv6 server.
+///
+/// Exposes internal fields and installs stub implementation of the
+/// @c CBControlDHCPv6 object.
+class NakedControlledDhcpv6Srv: public ControlledDhcpv6Srv {
+public:
+    /// @brief Constructor.
+    NakedControlledDhcpv6Srv()
+        : ControlledDhcpv6Srv(0) {
+        // We're replacing the @c CBControlDHCPv6 instance with our
+        // stub implementation used in tests.
+        cb_control_.reset(new TestCBControlDHCPv6());
+    }
+
     using ControlledDhcpv6Srv::signal_handler_;
 };
 
@@ -76,15 +181,105 @@ public:
     ///
     /// @param io_service Pointer to the IO service to be ran.
     /// @param timeout_ms Amount of time after which the method returns.
-    void runTimersWithTimeout(const IOServicePtr& io_service, const long timeout_ms) {
+    /// @param cond Pointer to the function which if returns true it
+    /// stops the IO service and causes the function to return.
+    void runTimersWithTimeout(const IOServicePtr& io_service, const long timeout_ms,
+                              std::function<bool()> cond = std::function<bool()>()) {
         IntervalTimer timer(*io_service);
-        timer.setup([&io_service]() {
+        bool stopped = false;
+        timer.setup([&io_service, &stopped]() {
             io_service->stop();
+            stopped = true;
         }, timeout_ms, IntervalTimer::ONE_SHOT);
-        io_service->run();
+
+        // Run as long as the timeout hasn't occurred and the interrupting
+        // condition is not specified or not met.
+        while (!stopped && (!cond || !cond())) {
+            io_service->run_one();
+        }
         io_service->get_io_service().reset();
     }
 
+    /// @brief This test verifies that the timer used to fetch the configuration
+    /// updates from the database works as expected.
+    void testConfigBackendTimer(const int config_wait_fetch_time,
+                                const bool throw_during_fetch = false) {
+        std::ostringstream config;
+        config <<
+            "{ \"Dhcp6\": {"
+            "\"interfaces-config\": {"
+            "    \"interfaces\": [ ]"
+            "},"
+            "\"lease-database\": {"
+            "     \"type\": \"memfile\","
+            "     \"persist\": false"
+            "},"
+            "\"config-control\": {"
+            "     \"config-fetch-wait-time\": " << config_wait_fetch_time <<
+            "},"
+            "\"rebind-timer\": 2000, "
+            "\"renew-timer\": 1000, \n"
+            "\"subnet6\": [ ],"
+            "\"valid-lifetime\": 4000 }"
+            "}";
+        writeFile(TEST_FILE, config.str());
+
+        // Create an instance of the server and initialize it.
+        boost::scoped_ptr<NakedControlledDhcpv6Srv> srv;
+        ASSERT_NO_THROW(srv.reset(new NakedControlledDhcpv6Srv()));
+        ASSERT_NO_THROW(srv->init(TEST_FILE));
+
+        // Get the CBControlDHCPv6 object belonging to this server.
+        auto cb_control = boost::dynamic_pointer_cast<TestCBControlDHCPv6>(srv->getCBControl());
+
+        // Verify that the parameter passed to the databaseConfigFetch has an
+        // expected value.
+        cb_control->enableCheckFetchMode();
+
+        // Instruct our stub implementation of the CBControlDHCPv6 to throw as a
+        // result of fetch if desired.
+        if (throw_during_fetch) {
+            cb_control->enableThrow();
+        }
+
+        // So far there should be exactly one attempt to fetch the configuration
+        // from the backend. That's the attempt made upon startup.
+        EXPECT_EQ(1, cb_control->getDatabaseConfigFetchCalls());
+
+
+        if ((config_wait_fetch_time > 0) && (!throw_during_fetch)) {
+            // If we're configured to run the timer, we expect that it was
+            // invoked at least 3 times. This is sufficient to verify that
+            // the timer was scheduled and that the timer continued to run
+            // even when an exception occurred during fetch (that's why it
+            // is 3 not 2).
+            ASSERT_NO_THROW(runTimersWithTimeout(srv->getIOService(), 500,
+                [cb_control]() {
+                    // Interrupt the timers poll if we have recorded at
+                    // least 3 attempts to fetch the updates.
+                    return (cb_control->getDatabaseConfigFetchCalls() >= 3);
+                }));
+            EXPECT_GE(cb_control->getDatabaseConfigFetchCalls(), 3);
+
+        } else {
+            ASSERT_NO_THROW(runTimersWithTimeout(srv->getIOService(), 500));
+
+            if (throw_during_fetch) {
+                // If we're simulating the failure condition the number
+                // of consecutive failures should not exceed 10. Therefore
+                // the number of recorded fetches should be 12. One at
+                // startup, 10 failures and one that causes the timer
+                // to stop.
+                EXPECT_EQ(12, cb_control->getDatabaseConfigFetchCalls());
+
+            } else {
+                // If the server is not configured to schedule the timer,
+                // we should still have one fetch attempt recorded.
+                EXPECT_EQ(1, cb_control->getDatabaseConfigFetchCalls());
+            }
+        }
+    }
+
     static const char* TEST_FILE;
     static const char* TEST_INCLUDE;
 };
@@ -626,6 +821,29 @@ TEST_F(JSONFileBackendTest, defaultLeaseDbBackend) {
     EXPECT_NO_THROW(static_cast<void>(LeaseMgrFactory::instance()));
 }
 
+
+// This test verifies that the timer triggering configuration updates
+// is invoked according to the configured value of the
+// config-fetch-wait-time.
+TEST_F(JSONFileBackendTest, configBackendTimer) {
+    testConfigBackendTimer(1);
+}
+
+// This test verifies that the timer for triggering configuration updates
+// is not invoked when the value of the config-fetch-wait-time is set
+// to 0.
+TEST_F(JSONFileBackendTest, configBackendTimerDisabled) {
+    testConfigBackendTimer(0);
+}
+
+// This test verifies that the server will gracefully handle exceptions
+// thrown from the CBControlDHCPv6::databaseConfigFetch, i.e. will
+// reschedule the timer.
+TEST_F(JSONFileBackendTest, configBackendTimerWithThrow) {
+    // The true value instructs the test to throw during the fetch.
+    testConfigBackendTimer(1, true);
+}
+
 // Starting tests which require MySQL backend availability. Those tests
 // will not be executed if Kea has been compiled without the
 // --with-mysql.
@@ -640,7 +858,7 @@ public:
     ///
     /// Recreates MySQL schema for a test.
     JSONFileBackendMySQLTest() : JSONFileBackendTest() {
-        destroyMySQLSchema();
+        // Ensure we have the proper schema with no transient data.
         createMySQLSchema();
     }
 
@@ -648,6 +866,7 @@ public:
     ///
     /// Destroys MySQL schema.
     virtual ~JSONFileBackendMySQLTest() {
+        // If data wipe enabled, delete transient data otherwise destroy the schema
         destroyMySQLSchema();
     }
 
