@@ -1,12 +1,14 @@
-// Copyright (C) 2016-2018 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2016-2019 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
 // file, You can obtain one at http://mozilla.org/MPL/2.0/.
 
+#include <config.h>
+
 #include <gtest/gtest.h>
-#include <cc/data.h>
 #include <dhcp6/parser_context.h>
+#include <dhcpsrv/parsers/simple_parser6.h>
 #include <testutils/io_utils.h>
 #include <testutils/user_context_utils.h>
 
@@ -268,6 +270,7 @@ void testFile(const std::string& fname) {
 TEST(ParserTest, file) {
     vector<string> configs;
     configs.push_back("advanced.json");
+    configs.push_back("all-keys-current.json");
     configs.push_back("backends.json");
     configs.push_back("cassandra.json");
     configs.push_back("classify.json");
@@ -275,6 +278,7 @@ TEST(ParserTest, file) {
     configs.push_back("comments.json");
     configs.push_back("dhcpv4-over-dhcpv6.json");
     configs.push_back("duid.json");
+    configs.push_back("global-reservations.json");
     configs.push_back("hooks.json");
     configs.push_back("iPXE.json");
     configs.push_back("leases-expiration.json");
@@ -294,6 +298,17 @@ TEST(ParserTest, file) {
     }
 }
 
+// This test loads the all-keys.json file and check global parameters.
+TEST(ParserTest, globalParameters) {
+    ConstElementPtr json;
+    Parser6Context ctx;
+    string fname = string(CFG_EXAMPLES) + "/" + "all-keys-current.json";
+    EXPECT_NO_THROW(json = ctx.parseFile(fname, Parser6Context::PARSER_DHCP6));
+    EXPECT_NO_THROW(json = json->get("Dhcp6"));
+    SimpleParser6 parser;
+    EXPECT_NO_THROW(parser.checkKeywords(parser.GLOBAL6_PARAMETERS, json));
+}
+
 /// @brief Tests error conditions in Dhcp6Parser
 ///
 /// @param txt text to be parsed
