@@ -1,4 +1,4 @@
-// Copyright (C) 2011-2017 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2011-2019 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -551,12 +551,10 @@ TEST_F(LibDhcpTest, unpackEmptyOption6) {
     LibDHCP::commitRuntimeOptionDefs();
 
     // Create the buffer holding the structure of the empty option.
-    const uint8_t raw_data[] = {
+    OptionBuffer buf = {
       0x04, 0x00,                // option code = 1024
       0x00, 0x00                 // option length = 0
     };
-    size_t raw_data_len = sizeof(raw_data) / sizeof(uint8_t);
-    OptionBuffer buf(raw_data, raw_data + raw_data_len);
 
     // Parse options.
     OptionCollection options;
@@ -599,7 +597,7 @@ TEST_F(LibDhcpTest, unpackSubOptions6) {
     LibDHCP::commitRuntimeOptionDefs();
 
     // Create the buffer holding the structure of options.
-    const char raw_data[] = {
+    OptionBuffer buf = {
         // First option starts here.
         0x00, 0x01,   // option code = 1
         0x00, 0x0F,   // option length = 15
@@ -613,7 +611,6 @@ TEST_F(LibDhcpTest, unpackSubOptions6) {
         0x00, 0x01,  // option length = 1
         0x00 // This option carries a single uint8 value and has no sub options.
     };
-    OptionBuffer buf(raw_data, raw_data + sizeof(raw_data));
 
     // Parse options.
     OptionCollection options;
@@ -919,12 +916,10 @@ TEST_F(LibDhcpTest, unpackEmptyOption4) {
     LibDHCP::commitRuntimeOptionDefs();
 
     // Create the buffer holding the structure of the empty option.
-    const uint8_t raw_data[] = {
+    OptionBuffer buf = {
       0xFE,                     // option code = 254
       0x00                      // option length = 0
     };
-    size_t raw_data_len = sizeof(raw_data) / sizeof(uint8_t);
-    OptionBuffer buf(raw_data, raw_data + raw_data_len);
 
     // Parse options.
     OptionCollection options;
@@ -969,7 +964,7 @@ TEST_F(LibDhcpTest, unpackSubOptions4) {
     LibDHCP::commitRuntimeOptionDefs();
 
     // Create the buffer holding the structure of options.
-    const uint8_t raw_data[] = {
+    OptionBuffer buf = {
         // First option starts here.
         0x01,                   // option code = 1
         0x0B,                   // option length = 11
@@ -984,8 +979,6 @@ TEST_F(LibDhcpTest, unpackSubOptions4) {
         0x00                    // This option carries a single uint8
                                 // value and has no sub options.
     };
-    size_t raw_data_len = sizeof(raw_data) / sizeof(uint8_t);
-    OptionBuffer buf(raw_data, raw_data + raw_data_len);
 
     // Parse options.
     OptionCollection options;
@@ -1016,6 +1009,126 @@ TEST_F(LibDhcpTest, unpackSubOptions4) {
     EXPECT_EQ(0x0, option_bar->getValue());
 }
 
+// Verifies that options 0 (PAD) and 255 (END) are handled as PAD and END
+// in and only in the dhcp4 space.
+TEST_F(LibDhcpTest, unpackPadEnd) {
+    // Create option definition for the container.
+    OptionDefinitionPtr opt_def(new OptionDefinition("container", 200,
+                                                     "empty", "my-space"));
+    // Create option definition for option 0.
+    OptionDefinitionPtr opt_def0(new OptionDefinition("zero", 0, "uint8"));
+
+    // Create option definition for option 255.
+    OptionDefinitionPtr opt_def255(new OptionDefinition("max", 255, "uint8"));
+
+    // Create option definition for another option.
+    OptionDefinitionPtr opt_def2(new OptionDefinition("my-option", 1,
+                                                      "string"));
+
+    // Register created option definitions as runtime option definitions.
+    OptionDefSpaceContainer defs;
+    ASSERT_NO_THROW(defs.addItem(opt_def, DHCP4_OPTION_SPACE));
+    ASSERT_NO_THROW(defs.addItem(opt_def0, "my-space"));
+    ASSERT_NO_THROW(defs.addItem(opt_def255, "my-space"));
+    ASSERT_NO_THROW(defs.addItem(opt_def2, "my-space"));
+    LibDHCP::setRuntimeOptionDefs(defs);
+    LibDHCP::commitRuntimeOptionDefs();
+
+    // Create the buffer holding the structure of options.
+    OptionBuffer buf = {
+        // Add a PAD
+        0x00,                         // option code = 0 (PAD)
+        // Container option starts here.
+        0xc8,                         // option code = 200 (container)
+        0x0b,                         // option length = 11
+        // Suboption 0.
+        0x00, 0x01, 0x00,             // code = 0, length = 1, content = 0
+        // Suboption 255.
+        0xff, 0x01, 0xff,             // code = 255, length = 1, content = 255
+        // Suboption 1.
+        0x01, 0x03, 0x66, 0x6f, 0x6f, // code = 1, length = 2, content = "foo"
+        // END
+        0xff,
+        // Extra bytes at tail.
+        0x01, 0x02, 0x03, 0x04
+    };
+
+    // Parse options.
+    OptionCollection options;
+    list<uint16_t> deferred;
+    size_t offset = 0;
+    ASSERT_NO_THROW(offset = LibDHCP::unpackOptions4(buf, DHCP4_OPTION_SPACE,
+                                                     options, deferred));
+
+    // Returned offset should point to the END.
+    EXPECT_EQ(0xff, buf[offset]);
+
+    // There should be one top level option.
+    ASSERT_EQ(1, options.size());
+
+    // Get it.
+    OptionPtr option = options.begin()->second;
+    ASSERT_TRUE(option);
+    EXPECT_EQ(200, option->getType());
+
+    // There should be 3 suboptions.
+    ASSERT_EQ(3, option->getOptions().size());
+
+    // Get suboption 0.
+    boost::shared_ptr<OptionInt<uint8_t> > sub0 =
+        boost::dynamic_pointer_cast<OptionInt<uint8_t> >
+            (option->getOption(0));
+    ASSERT_TRUE(sub0);
+    EXPECT_EQ(0, sub0->getType());
+    EXPECT_EQ(0, sub0->getValue());
+
+    // Get suboption 255.
+    boost::shared_ptr<OptionInt<uint8_t> > sub255 =
+        boost::dynamic_pointer_cast<OptionInt<uint8_t> >
+            (option->getOption(255));
+    ASSERT_TRUE(sub255);
+    EXPECT_EQ(255, sub255->getType());
+    EXPECT_EQ(255, sub255->getValue());
+
+    // Get suboption 1.
+    boost::shared_ptr<OptionString> sub =
+        boost::dynamic_pointer_cast<OptionString>(option->getOption(1));
+    ASSERT_TRUE(sub);
+    EXPECT_EQ(1, sub->getType());
+    EXPECT_EQ("foo", sub->getValue());
+}
+
+// Verifies that an Host Name (option 12), will be dropped when empty,
+// while subsequent options will still be unpacked.
+TEST_F(LibDhcpTest, emptyHostName) {
+
+    uint8_t opts[] = {
+        12,  0,             // Empty Hostname
+        60,  3, 10, 11, 12  // Class Id
+    };
+
+    vector<uint8_t> packed(opts, opts + sizeof(opts));
+    isc::dhcp::OptionCollection options; // list of options
+    list<uint16_t> deferred;
+
+    ASSERT_NO_THROW(
+        LibDHCP::unpackOptions4(packed, "dhcp4", options, deferred);
+    );
+
+    // Host Name should not exist, we quietly drop it when empty.
+    isc::dhcp::OptionCollection::const_iterator x = options.find(12);
+    ASSERT_TRUE(x == options.end());
+
+    // Verify Option 60 exists correctly
+    x = options.find(60);
+    ASSERT_FALSE(x == options.end());
+    EXPECT_EQ(60, x->second->getType());
+    ASSERT_EQ(3, x->second->getData().size());
+    EXPECT_EQ(5, x->second->len());
+    EXPECT_EQ(0, memcmp(&x->second->getData()[0], opts + 4, 3));
+};
+
+
 TEST_F(LibDhcpTest, stdOptionDefs4) {
 
     // Create a buffer that holds dummy option data.
@@ -1742,18 +1855,6 @@ TEST_F(LibDhcpTest, stdOptionDefs6) {
     LibDhcpTest::testStdOptionDefs6(D6O_IPV6_ADDRESS_ANDSF, begin, end,
                                     typeid(Option6AddrLst));
 
-    LibDhcpTest::testStdOptionDefs6(D6O_PUBLIC_KEY, begin, end,
-                                    typeid(Option));
-
-    LibDhcpTest::testStdOptionDefs6(D6O_CERTIFICATE, begin, end,
-                                    typeid(Option));
-
-    LibDhcpTest::testStdOptionDefs6(D6O_SIGNATURE, begin, end,
-                                    typeid(OptionCustom));
-
-    LibDhcpTest::testStdOptionDefs6(D6O_TIMESTAMP, begin, begin + 8,
-                                    typeid(Option));
-
     // RFC7598 options
     LibDhcpTest::testOptionDefs6(MAPE_V6_OPTION_SPACE, D6O_S46_RULE, begin, end,
                                  typeid(OptionCustom), "s46-rule-options");
