@@ -1,4 +1,4 @@
-// Copyright (C) 2012-2017 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2012-2019 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -7,16 +7,19 @@
 #include <config.h>
 
 #include <asiolink/io_address.h>
+#include <asiolink/addr_utilities.h>
 #include <dhcp/option_space.h>
-#include <dhcpsrv/addr_utilities.h>
 #include <dhcpsrv/shared_network.h>
 #include <dhcpsrv/subnet.h>
+#include <boost/lexical_cast.hpp>
+#include <boost/make_shared.hpp>
 #include <algorithm>
 #include <sstream>
 
 using namespace isc::asiolink;
 using namespace isc::data;
 using namespace isc::dhcp;
+using namespace isc::util;
 
 namespace {
 
@@ -56,12 +59,21 @@ Subnet::Subnet(const isc::asiolink::IOAddress& prefix, uint8_t len,
       prefix_len_(len),
       last_allocated_ia_(lastAddrInPrefix(prefix, len)),
       last_allocated_ta_(lastAddrInPrefix(prefix, len)),
-      last_allocated_pd_(lastAddrInPrefix(prefix, len)) {
+      last_allocated_pd_(lastAddrInPrefix(prefix, len)),
+      last_allocated_time_(),
+      iface_(),
+      shared_network_name_() {
     if ((prefix.isV6() && len > 128) ||
         (prefix.isV4() && len > 32)) {
         isc_throw(BadValue,
                   "Invalid prefix length specified for subnet: " << len);
     }
+
+    // Initialize timestamps for each lease type to negative infinity.
+    last_allocated_time_[Lease::TYPE_V4] = boost::posix_time::neg_infin;
+    last_allocated_time_[Lease::TYPE_NA] = boost::posix_time::neg_infin;
+    last_allocated_time_[Lease::TYPE_TA] = boost::posix_time::neg_infin;
+    last_allocated_time_[Lease::TYPE_PD] = boost::posix_time::neg_infin;
 }
 
 bool
@@ -89,6 +101,19 @@ isc::asiolink::IOAddress Subnet::getLastAllocated(Lease::Type type) const {
     }
 }
 
+boost::posix_time::ptime
+Subnet::getLastAllocatedTime(const Lease::Type& lease_type) const {
+    auto t = last_allocated_time_.find(lease_type);
+    if (t != last_allocated_time_.end()) {
+        return (t->second);
+    }
+
+    // This shouldn't happen, because we have initialized the structure
+    // for all lease types.
+    return (boost::posix_time::neg_infin);
+}
+
+
 void Subnet::setLastAllocated(Lease::Type type,
                               const isc::asiolink::IOAddress& addr) {
 
@@ -99,16 +124,19 @@ void Subnet::setLastAllocated(Lease::Type type,
     case Lease::TYPE_V4:
     case Lease::TYPE_NA:
         last_allocated_ia_ = addr;
-        return;
+        break;
     case Lease::TYPE_TA:
         last_allocated_ta_ = addr;
-        return;
+        break;
     case Lease::TYPE_PD:
         last_allocated_pd_ = addr;
-        return;
+        break;
     default:
         isc_throw(BadValue, "Pool type " << type << " not supported");
     }
+
+    // Update the timestamp of last allocation.
+    last_allocated_time_[type] = boost::posix_time::microsec_clock::universal_time();
 }
 
 std::string
@@ -134,6 +162,23 @@ Subnet::getPoolCapacity(Lease::Type type) const {
     }
 }
 
+uint64_t
+Subnet::getPoolCapacity(Lease::Type type,
+                        const ClientClasses& client_classes) const {
+    switch (type) {
+    case Lease::TYPE_V4:
+    case Lease::TYPE_NA:
+        return sumPoolCapacity(pools_, client_classes);
+    case Lease::TYPE_TA:
+        return sumPoolCapacity(pools_ta_, client_classes);
+    case Lease::TYPE_PD:
+        return sumPoolCapacity(pools_pd_, client_classes);
+    default:
+        isc_throw(BadValue, "Unsupported pool type: "
+                  << static_cast<int>(type));
+    }
+}
+
 uint64_t
 Subnet::sumPoolCapacity(const PoolCollection& pools) const {
     uint64_t sum = 0;
@@ -152,31 +197,82 @@ Subnet::sumPoolCapacity(const PoolCollection& pools) const {
     return (sum);
 }
 
+uint64_t
+Subnet::sumPoolCapacity(const PoolCollection& pools,
+                        const ClientClasses& client_classes) const {
+    uint64_t sum = 0;
+    for (PoolCollection::const_iterator p = pools.begin(); p != pools.end(); ++p) {
+        if (!(*p)->clientSupported(client_classes)) {
+            continue;
+        }
+        uint64_t x = (*p)->getCapacity();
+
+        // Check if we can add it. If sum + x > uint64::max, then we would have
+        // overflown if we tried to add it.
+        if (x > std::numeric_limits<uint64_t>::max() - sum) {
+            return (std::numeric_limits<uint64_t>::max());
+        }
+
+        sum += x;
+    }
+
+    return (sum);
+}
+
+std::pair<IOAddress, uint8_t>
+Subnet::parsePrefixCommon(const std::string& prefix) {
+    auto pos = prefix.find('/');
+    if ((pos == std::string::npos) ||
+        (pos == prefix.size() - 1) ||
+        (pos == 0)) {
+        isc_throw(BadValue, "unable to parse invalid prefix " << prefix);
+    }
+
+    try {
+        IOAddress address(prefix.substr(0, pos));
+        int length = boost::lexical_cast<int>(prefix.substr(pos + 1));
+        return (std::make_pair(address, static_cast<int>(length)));
+
+    } catch (...) {
+        isc_throw(BadValue, "unable to parse invalid prefix " << prefix);
+    }
+}
+
+
 void Subnet4::checkType(Lease::Type type) const {
     if (type != Lease::TYPE_V4) {
         isc_throw(BadValue, "Only TYPE_V4 is allowed for Subnet4");
     }
 }
 
-Subnet4::Subnet4(const isc::asiolink::IOAddress& prefix, uint8_t length,
+Subnet4::Subnet4(const IOAddress& prefix, uint8_t length,
                  const Triplet<uint32_t>& t1,
                  const Triplet<uint32_t>& t2,
                  const Triplet<uint32_t>& valid_lifetime,
                  const SubnetID id)
-    : Subnet(prefix, length, id), Network4(),
-      siaddr_(IOAddress("0.0.0.0")) {
+    : Subnet(prefix, length, id), Network4() {
     if (!prefix.isV4()) {
         isc_throw(BadValue, "Non IPv4 prefix " << prefix.toText()
                   << " specified in subnet4");
     }
-    // Relay info.
-    setRelayInfo(IOAddress::IPV4_ZERO_ADDRESS());
+
     // Timers.
     setT1(t1);
     setT2(t2);
     setValid(valid_lifetime);
 }
 
+Subnet4Ptr
+Subnet4::create(const IOAddress& prefix, uint8_t length,
+                const Triplet<uint32_t>& t1,
+                const Triplet<uint32_t>& t2,
+                const Triplet<uint32_t>& valid_lifetime,
+                const SubnetID id) {
+    Subnet4Ptr subnet = boost::make_shared<Subnet4>
+        (prefix, length, t1, t2, valid_lifetime, id);
+    return (subnet);
+}
+
 Subnet4Ptr
 Subnet4::getNextSubnet(const Subnet4Ptr& first_subnet) const {
     SharedNetwork4Ptr network;
@@ -227,18 +323,6 @@ Subnet4::clientSupported(const isc::dhcp::ClientClasses& client_classes) const {
     return (Network4::clientSupported(client_classes));
 }
 
-void Subnet4::setSiaddr(const isc::asiolink::IOAddress& siaddr) {
-    if (!siaddr.isV4()) {
-        isc_throw(BadValue, "Can't set siaddr to non-IPv4 address "
-                  << siaddr);
-    }
-    siaddr_ = siaddr;
-}
-
-isc::asiolink::IOAddress Subnet4::getSiaddr() const {
-    return (siaddr_);
-}
-
 const PoolCollection& Subnet::getPools(Lease::Type type) const {
     // check if the type is valid (and throw if it isn't)
     checkType(type);
@@ -314,6 +398,34 @@ const PoolPtr Subnet::getPool(Lease::Type type, const isc::asiolink::IOAddress&
     return (candidate);
 }
 
+const PoolPtr Subnet::getPool(Lease::Type type,
+                              const ClientClasses& client_classes,
+                              const isc::asiolink::IOAddress& hint) const {
+    // check if the type is valid (and throw if it isn't)
+    checkType(type);
+
+    const PoolCollection& pools = getPools(type);
+
+    PoolPtr candidate;
+
+    if (!pools.empty()) {
+        PoolCollection::const_iterator ub =
+            std::upper_bound(pools.begin(), pools.end(), hint,
+                             prefixLessThanFirstAddress);
+
+        if (ub != pools.begin()) {
+            --ub;
+            if ((*ub)->inRange(hint) &&
+                (*ub)->clientSupported(client_classes)) {
+                candidate = *ub;
+            }
+        }
+    }
+
+    // Return a pool or NULL if no match found.
+    return (candidate);
+}
+
 void
 Subnet::addPool(const PoolPtr& pool) {
     // check if the type is valid (and throw if it isn't)
@@ -394,6 +506,31 @@ Subnet::inPool(Lease::Type type, const isc::asiolink::IOAddress& addr) const {
     return (false);
 }
 
+bool
+Subnet::inPool(Lease::Type type,
+               const isc::asiolink::IOAddress& addr,
+               const ClientClasses& client_classes) const {
+
+    // Let's start with checking if it even belongs to that subnet.
+    if ((type != Lease::TYPE_PD) && !inRange(addr)) {
+        return (false);
+    }
+
+    const PoolCollection& pools = getPools(type);
+
+    for (PoolCollection::const_iterator pool = pools.begin();
+         pool != pools.end(); ++pool) {
+        if (!(*pool)->clientSupported(client_classes)) {
+            continue;
+        }
+        if ((*pool)->inRange(addr)) {
+            return (true);
+        }
+    }
+    // There's no pool that address belongs to
+    return (false);
+}
+
 bool
 Subnet::poolOverlaps(const Lease::Type& pool_type, const PoolPtr& pool) const {
     const PoolCollection& pools = getPools(pool_type);
@@ -460,7 +597,7 @@ Subnet::poolOverlaps(const Lease::Type& pool_type, const PoolPtr& pool) const {
 }
 
 
-Subnet6::Subnet6(const isc::asiolink::IOAddress& prefix, uint8_t length,
+Subnet6::Subnet6(const IOAddress& prefix, uint8_t length,
                  const Triplet<uint32_t>& t1,
                  const Triplet<uint32_t>& t2,
                  const Triplet<uint32_t>& preferred_lifetime,
@@ -472,8 +609,6 @@ Subnet6::Subnet6(const isc::asiolink::IOAddress& prefix, uint8_t length,
                   << " specified in subnet6");
     }
 
-    // Relay info.
-    setRelayInfo(RelayInfo(IOAddress::IPV6_ZERO_ADDRESS()));
     // Timers.
     setT1(t1);
     setT2(t2);
@@ -481,6 +616,18 @@ Subnet6::Subnet6(const isc::asiolink::IOAddress& prefix, uint8_t length,
     setValid(valid_lifetime);
 }
 
+Subnet6Ptr
+Subnet6::create(const IOAddress& prefix, uint8_t length,
+                const Triplet<uint32_t>& t1,
+                const Triplet<uint32_t>& t2,
+                const Triplet<uint32_t>& preferred_lifetime,
+                const Triplet<uint32_t>& valid_lifetime,
+                const SubnetID id) {
+    Subnet6Ptr subnet = boost::make_shared<Subnet6>
+        (prefix, length, t1, t2, preferred_lifetime, valid_lifetime, id);
+    return (subnet);
+}
+
 void Subnet6::checkType(Lease::Type type) const {
     if ( (type != Lease::TYPE_NA) && (type != Lease::TYPE_TA) &&
          (type != Lease::TYPE_PD)) {
@@ -543,6 +690,9 @@ data::ElementPtr
 Subnet::toElement() const {
     ElementPtr map = Element::createMap();
 
+    // Add user-context
+    contextToElement(map);
+
     // Set subnet id
     SubnetID id = getID();
     map->set("id", Element::create(static_cast<long long>(id)));
@@ -550,12 +700,6 @@ Subnet::toElement() const {
     // Set subnet
     map->set("subnet", Element::create(toText()));
 
-    // Add user-context, but only if defined. Omit if it was not.
-    ConstElementPtr ctx = getContext();
-    if (ctx) {
-        map->set("user-context", ctx);
-    }
-
     return (map);
 }
 
@@ -571,9 +715,6 @@ Subnet4::toElement() const {
     const Cfg4o6& d4o6 = get4o6();
     isc::data::merge(map, d4o6.toElement());
 
-    // Set next-server
-    map->set("next-server", Element::create(getSiaddr().toText()));
-
     // Set pools
     const PoolCollection& pools = getPools(Lease::TYPE_V4);
     ElementPtr pool_list = Element::createList();
@@ -587,6 +728,16 @@ Subnet4::toElement() const {
     return (map);
 }
 
+std::pair<IOAddress, uint8_t>
+Subnet4::parsePrefix(const std::string& prefix) {
+    std::pair<IOAddress, uint8_t> parsed = Subnet::parsePrefixCommon(prefix);
+    if (!parsed.first.isV4() || parsed.first.isV4Zero() ||
+        (parsed.second > 32) || (parsed.second == 0)) {
+        isc_throw(BadValue, "unable to parse invalid IPv4 prefix " << prefix);
+    }
+    return (parsed);
+}
+
 data::ElementPtr
 Subnet6::toElement() const {
     // Prepare the map
@@ -600,90 +751,34 @@ Subnet6::toElement() const {
     ElementPtr pool_list = Element::createList();
     for (PoolCollection::const_iterator pool = pools.cbegin();
          pool != pools.cend(); ++pool) {
-        // Prepare the map for a pool (@todo move this code to pool.cc)
-        ElementPtr pool_map = Element::createMap();
-        // Set pool
-        const IOAddress& first = (*pool)->getFirstAddress();
-        const IOAddress& last = (*pool)->getLastAddress();
-        std::string range = first.toText() + "-" + last.toText();
-        // Try to output a prefix (vs a range)
-        int prefix_len = prefixLengthFromRange(first, last);
-        if (prefix_len >= 0) {
-            std::ostringstream oss;
-            oss << first.toText() << "/" << prefix_len;
-            range = oss.str();
-        }
-        pool_map->set("pool", Element::create(range));
-        // Set user-context
-        ConstElementPtr context = (*pool)->getContext();
-        if (!isNull(context)) {
-            pool_map->set("user-context", context);
-        }
-        // Set pool options
-        ConstCfgOptionPtr opts = (*pool)->getCfgOption();
-        pool_map->set("option-data", opts->toElement());
-        // Push on the pool list
-        pool_list->add(pool_map);
+        // Add the elementized pool to the list
+        pool_list->add((*pool)->toElement());
     }
     map->set("pools", pool_list);
+
     // Set pd-pools
     const PoolCollection& pdpools = getPools(Lease::TYPE_PD);
     ElementPtr pdpool_list = Element::createList();
     for (PoolCollection::const_iterator pool = pdpools.cbegin();
          pool != pdpools.cend(); ++pool) {
-        // Get it as a Pool6 (@todo move this code to pool.cc)
-        const Pool6* pdpool = dynamic_cast<Pool6*>(pool->get());
-        if (!pdpool) {
-            isc_throw(ToElementError, "invalid pd-pool pointer");
-        }
-        // Prepare the map for a pd-pool
-        ElementPtr pool_map = Element::createMap();
-        // Set prefix
-        const IOAddress& prefix = pdpool->getFirstAddress();
-        pool_map->set("prefix", Element::create(prefix.toText()));
-        // Set prefix-len (get it from min - max)
-        const IOAddress& last = pdpool->getLastAddress();
-        int prefix_len = prefixLengthFromRange(prefix, last);
-        if (prefix_len < 0) {
-            // The pool is bad: give up
-            isc_throw(ToElementError, "invalid prefix range "
-                      << prefix.toText() << "-" << last.toText());
-        }
-        pool_map->set("prefix-len", Element::create(prefix_len));
-        // Set delegated-len
-        uint8_t len = pdpool->getLength();
-        pool_map->set("delegated-len",
-                      Element::create(static_cast<int>(len)));
-
-        // Set excluded prefix
-        const Option6PDExcludePtr& xopt =
-            pdpool->getPrefixExcludeOption();
-        if (xopt) {
-            const IOAddress& xprefix =
-                xopt->getExcludedPrefix(prefix, len);
-            pool_map->set("excluded-prefix",
-                          Element::create(xprefix.toText()));
-            uint8_t xlen = xopt->getExcludedPrefixLength();
-            pool_map->set("excluded-prefix-len",
-                          Element::create(static_cast<int>(xlen)));
-        }
-
-        // Set user-context
-        ConstElementPtr context = pdpool->getContext();
-        if (!isNull(context)) {
-            pool_map->set("user-context", context);
-        }
-        // Set pool options
-        ConstCfgOptionPtr opts = pdpool->getCfgOption();
-        pool_map->set("option-data", opts->toElement());
-        // Push on the pool list
-        pdpool_list->add(pool_map);
+        // Add the elementized pool to the list
+        pdpool_list->add((*pool)->toElement());
     }
     map->set("pd-pools", pdpool_list);
 
     return (map);
 }
 
+std::pair<IOAddress, uint8_t>
+Subnet6::parsePrefix(const std::string& prefix) {
+    std::pair<IOAddress, uint8_t> parsed = Subnet::parsePrefixCommon(prefix);
+    if (!parsed.first.isV6() || parsed.first.isV6Zero() ||
+        (parsed.second > 128) || (parsed.second == 0)) {
+        isc_throw(BadValue, "unable to parse invalid IPv6 prefix " << prefix);
+    }
+    return (parsed);
+}
+
 
 } // end of isc::dhcp namespace
 } // end of isc namespace
