@@ -1,4 +1,4 @@
-// Copyright (C) 2014-2017 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2014-2019 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -175,33 +175,174 @@ TEST_F(CfgHostsTest, getAllRepeatingHosts) {
 
     // Verify that hosts can be retrieved.
     for (unsigned i = 0; i < 25; ++i) {
-        // Get host by HW address. The DUID is non-null but the reservation
-        // should be returned for the HW address because there are no
-        // reservations for the DUIDs from the range of 25 to 49.
-        HostCollection hosts = cfg.getAll(hwaddrs_[i], duids_[i + 25]);
+        // Get host by HW address.
+        HostCollection hosts = cfg.getAll(Host::IDENT_HWADDR,
+                                          &hwaddrs_[i]->hwaddr_[0],
+                                          hwaddrs_[i]->hwaddr_.size());
         ASSERT_EQ(2, hosts.size());
         EXPECT_EQ(1, hosts[0]->getIPv4SubnetID());
         EXPECT_EQ(addressesa_[i], hosts[0]->getIPv4Reservation().toText());
         EXPECT_EQ(2, hosts[1]->getIPv4SubnetID());
         EXPECT_EQ(addressesb_[i], hosts[1]->getIPv4Reservation().toText());
 
-        // Get host by DUID. The HW address is non-null but the reservation
-        // should be returned for the DUID because there are no
-        // reservations for the HW addresses from the range of 25 to 49.
-        hosts = cfg.getAll(hwaddrs_[i + 25], duids_[i]);
-        ASSERT_EQ(2, hosts.size());
-        EXPECT_EQ(1, hosts[0]->getIPv4SubnetID());
-        EXPECT_EQ(2, hosts[1]->getIPv4SubnetID());
+        // The HW address is non-null but there are no reservations
+        // for the HW addresses from the range of 25 to 49.
+        hosts = cfg.getAll(Host::IDENT_HWADDR,
+                           &hwaddrs_[i + 25]->hwaddr_[0],
+                           hwaddrs_[i + 25]->hwaddr_.size());
+        EXPECT_TRUE(hosts.empty());
+
+        // Get host by DUID.
+        hosts = cfg.getAll(Host::IDENT_DUID,
+                           &duids_[i]->getDuid()[0],
+                           duids_[i]->getDuid().size());
+
+        // The DUID is non-null but there are no reservations
+        // for the DUIDs from the range of 25 to 49.
+        hosts = cfg.getAll(Host::IDENT_DUID,
+                           &duids_[i + 25]->getDuid()[0],
+                           duids_[i + 25]->getDuid().size());
+        EXPECT_TRUE(hosts.empty());
+    }
+}
+
+// This test checks that hosts in the same subnet can be retrieved from
+// the host configuration.
+TEST_F(CfgHostsTest, getAll4BySubnet) {
+    CfgHosts cfg;
+    // Add 25 hosts identified by HW address in the same subnet.
+    for (unsigned i = 0; i < 25; ++i) {
+        cfg.add(HostPtr(new Host(hwaddrs_[i]->toText(false),
+                                 "hw-address",
+                                 SubnetID(1), SubnetID(1),
+                                 addressesa_[i])));
     }
 
-    // The getAll function should return empty containers for the HW addresses
-    //  and DUIDs for which the reservations haven't been added.
-    for (int i = 25; i < 50; ++i) {
-        EXPECT_TRUE(cfg.getAll(hwaddrs_[i]).empty());
-        EXPECT_TRUE(cfg.getAll(HWAddrPtr(), duids_[i]).empty());
+    // Check that other subnets are empty.
+    HostCollection hosts = cfg.getAll4(SubnetID(100));
+    EXPECT_EQ(0, hosts.size());
+
+    // Try to retrieve all added reservations.
+    hosts = cfg.getAll4(SubnetID(1));
+    ASSERT_EQ(25, hosts.size());
+    for (unsigned i = 0; i < 25; ++i) {
+        EXPECT_EQ(1, hosts[i]->getIPv4SubnetID());
+        EXPECT_EQ(addressesa_[i].toText(),
+                  hosts[i]->getIPv4Reservation().toText());
     }
 }
 
+// This test checks that hosts in the same subnet can be retrieved from
+// the host configuration.
+TEST_F(CfgHostsTest, getAll6BySubnet) {
+    CfgHosts cfg;
+    // Add 25 hosts identified by DUID in the same subnet.
+    for (unsigned i = 0; i < 25; ++i) {
+        HostPtr host = HostPtr(new Host(duids_[i]->toText(), "duid",
+                                        SubnetID(1), SubnetID(1),
+                                        IOAddress("0.0.0.0")));
+        host->addReservation(IPv6Resrv(IPv6Resrv::TYPE_NA,
+                                       increase(IOAddress("2001:db8:1::1"),
+                                                i)));
+        cfg.add(host);
+    }
+
+    // Check that other subnets are empty.
+    HostCollection hosts = cfg.getAll6(SubnetID(100));
+    EXPECT_EQ(0, hosts.size());
+
+    // Try to retrieve all added reservations.
+    hosts = cfg.getAll6(SubnetID(1));
+    ASSERT_EQ(25, hosts.size());
+    for (unsigned i = 0; i < 25; ++i) {
+        EXPECT_EQ(1, hosts[i]->getIPv6SubnetID());
+        IPv6ResrvRange reservations =
+            hosts[i]->getIPv6Reservations(IPv6Resrv::TYPE_NA);
+        ASSERT_EQ(1, std::distance(reservations.first, reservations.second));
+        EXPECT_EQ(increase(IOAddress("2001:db8:1::1"), i),
+                  reservations.first->second.getPrefix());
+    }
+}
+
+// This test checks that hosts in the same subnet can be retrieved from
+// the host configuration by pages.
+TEST_F(CfgHostsTest, getPage4) {
+    CfgHosts cfg;
+    // Add 25 hosts identified by DUID in the same subnet.
+    for (unsigned i = 0; i < 25; ++i) {
+        cfg.add(HostPtr(new Host(duids_[i]->toText(), "duid",
+                                 SubnetID(1), SubnetID(1),
+                                 addressesa_[i])));
+    }
+    size_t idx(0);
+    uint64_t host_id(0);
+    HostPageSize page_size(10);
+
+    // Check that other subnets are empty.
+    HostCollection page = cfg.getPage4(SubnetID(100), idx, host_id, page_size);
+    EXPECT_EQ(0, page.size());
+
+    // Try to retrieve all added reservations.
+    // Get first page.
+    page = cfg.getPage4(SubnetID(1), idx, host_id, page_size);
+    EXPECT_EQ(10, page.size());
+    host_id = page[9]->getHostId();
+
+    // Get second and last pages.
+    page = cfg.getPage4(SubnetID(1), idx, host_id, page_size);
+    EXPECT_EQ(10, page.size());
+    host_id = page[9]->getHostId();
+    page = cfg.getPage4(SubnetID(1), idx, host_id, page_size);
+    EXPECT_EQ(5, page.size());
+    host_id = page[4]->getHostId();
+
+    // Verify we have everything.
+    page = cfg.getPage4(SubnetID(1), idx, host_id, page_size);
+    EXPECT_EQ(0, page.size());
+}
+
+// This test checks that hosts in the same subnet can be retrieved from
+// the host configuration by pages.
+TEST_F(CfgHostsTest, getPage6) {
+    CfgHosts cfg;
+    // Add 25 hosts identified by HW address in the same subnet.
+    for (unsigned i = 0; i < 25; ++i) {
+        HostPtr host = HostPtr(new Host(hwaddrs_[i]->toText(false),
+                                        "hw-address",
+                                        SubnetID(1), SubnetID(1),
+                                        IOAddress("0.0.0.0")));
+        host->addReservation(IPv6Resrv(IPv6Resrv::TYPE_NA,
+                                       increase(IOAddress("2001:db8:1::1"),
+                                                i)));
+        cfg.add(host);
+    }
+    size_t idx(0);
+    uint64_t host_id(0);
+    HostPageSize page_size(10);
+
+    // Check that other subnets are empty.
+    HostCollection page = cfg.getPage6(SubnetID(100), idx, host_id, page_size);
+    EXPECT_EQ(0, page.size());
+
+    // Try to retrieve all added reservations.
+    // Get first page.
+    page = cfg.getPage6(SubnetID(1), idx, host_id, page_size);
+    EXPECT_EQ(10, page.size());
+    host_id = page[9]->getHostId();
+
+    // Get second and last pages.
+    page = cfg.getPage6(SubnetID(1), idx, host_id, page_size);
+    EXPECT_EQ(10, page.size());
+    host_id = page[9]->getHostId();
+    page = cfg.getPage6(SubnetID(1), idx, host_id, page_size);
+    EXPECT_EQ(5, page.size());
+    host_id = page[4]->getHostId();
+
+    // Verify we have everything.
+    page = cfg.getPage6(SubnetID(1), idx, host_id, page_size);
+    EXPECT_EQ(0, page.size());
+}
+
 // This test checks that all reservations for the specified IPv4 address can
 // be retrieved.
 TEST_F(CfgHostsTest, getAll4ByAddress) {
@@ -211,12 +352,12 @@ TEST_F(CfgHostsTest, getAll4ByAddress) {
         // Add host identified by the HW address.
         cfg.add(HostPtr(new Host(hwaddrs_[i]->toText(false),
                                  "hw-address",
-                                 SubnetID(1 + i), SubnetID(0),
+                                 SubnetID(1 + i), SUBNET_ID_UNUSED,
                                  IOAddress("192.0.2.5"))));
         // Add host identified by the DUID.
         cfg.add(HostPtr(new Host(duids_[i]->toText(),
                                  "duid",
-                                 SubnetID(1 + i), SubnetID(0),
+                                 SubnetID(1 + i), SUBNET_ID_UNUSED,
                                  IOAddress("192.0.2.10"))));
     }
 
@@ -231,6 +372,54 @@ TEST_F(CfgHostsTest, getAll4ByAddress) {
     EXPECT_EQ(25, *subnet_ids.rbegin());
 }
 
+// This test checks that all reservations for the specified IPv4 subnet can
+// be deleted.
+TEST_F(CfgHostsTest, deleteAll4) {
+    CfgHosts cfg;
+    // Add hosts.
+    for (unsigned i = 0; i < 25; ++i) {
+        // Hosts will differ by hostname. It is easier than differentiating by
+        // IPv4 address because if they all have zero IPv4 address it is
+        // easier to retrieve all of them to check the host counts.
+        std::ostringstream s;
+        s << "hostname" << i;
+
+        // Add host identified by the HW address.
+        cfg.add(HostPtr(new Host(hwaddrs_[i]->toText(false),
+                                 "hw-address",
+                                 SubnetID(1 + i % 2), SUBNET_ID_UNUSED,
+                                 IOAddress::IPV4_ZERO_ADDRESS(),
+                                 "hostname")));
+    }
+
+    // Get all inserted hosts.
+    HostCollection hosts = cfg.getAll4(IOAddress::IPV4_ZERO_ADDRESS());
+    std::set<uint32_t> subnet_ids;
+    for (HostCollection::const_iterator host = hosts.begin(); host != hosts.end();
+         ++host) {
+        subnet_ids.insert((*host)->getIPv4SubnetID());
+    }
+    // Make sure there are two unique subnets: 1 and 2.
+    ASSERT_EQ(2, subnet_ids.size());
+    EXPECT_EQ(1, *subnet_ids.begin());
+    EXPECT_EQ(2, *subnet_ids.rbegin());
+
+    // Delete all hosts for subnet id 2. There should be 12 of them.
+    EXPECT_EQ(12, cfg.delAll4(SubnetID(2)));
+
+    // Gather the host counts again.
+    subnet_ids.clear();
+    hosts = cfg.getAll4(IOAddress::IPV4_ZERO_ADDRESS());
+    for (HostCollection::const_iterator host = hosts.begin(); host != hosts.end();
+         ++host) {
+        subnet_ids.insert((*host)->getIPv4SubnetID());
+    }
+    // We should only have hosts for one subnet and it should be the subnet
+    // with ID of 1.
+    ASSERT_EQ(1, subnet_ids.size());
+    EXPECT_EQ(1, *subnet_ids.begin());
+}
+
 // This test checks that the reservations can be retrieved for the particular
 // host connected to the specific IPv4 subnet (by subnet id).
 TEST_F(CfgHostsTest, get4) {
@@ -405,6 +594,53 @@ TEST_F(CfgHostsTest, get6) {
     }
 }
 
+// This test checks that all reservations for the specified IPv6 subnet can
+// be deleted.
+TEST_F(CfgHostsTest, deleteAll6) {
+    CfgHosts cfg;
+    // Add hosts.
+    for (unsigned i = 0; i < 25; ++i) {
+        // Add host identified by HW address. The subnet for which we're
+        // adding the host has id of 1 for even values of i and 2 for
+        // odd values of i.
+        HostPtr host = HostPtr(new Host(hwaddrs_[i]->toText(false),
+                                        "hw-address",
+                                        SubnetID(10), SubnetID(1 + i % 2),
+                                        IOAddress("0.0.0.0")));
+        host->addReservation(IPv6Resrv(IPv6Resrv::TYPE_NA,
+                                       increase(IOAddress("2001:db8:1::1"),
+                                                i)));
+        cfg.add(host);
+    }
+
+    // Delete all hosts for subnet id. There should be 13 of them.
+    EXPECT_EQ(13, cfg.delAll6(SubnetID(1)));
+
+    for (unsigned i = 0; i < 25; ++i) {
+        // Calculate subnet id for the given i.
+        SubnetID subnet_id = 1 + i % 2;
+
+        // Try to retrieve host by HW address.
+        HostPtr host = cfg.get6(subnet_id, Host::IDENT_HWADDR,
+                                &hwaddrs_[i]->hwaddr_[0],
+                                hwaddrs_[i]->hwaddr_.size());
+        // The host should exist for subnet id of 2.
+        if (subnet_id == 2) {
+            ASSERT_TRUE(host);
+            EXPECT_EQ(subnet_id, host->getIPv6SubnetID());
+            IPv6ResrvRange reservations =
+                host->getIPv6Reservations(IPv6Resrv::TYPE_NA);
+            ASSERT_EQ(1, std::distance(reservations.first, reservations.second));
+            EXPECT_EQ(increase(IOAddress("2001:db8:1::1"), i),
+                      reservations.first->second.getPrefix());
+
+        } else {
+            // All hosts for subnet id 2 should be gone.
+            EXPECT_FALSE(host);
+        }
+    }
+}
+
 // This test checks that the DHCPv6 reservations can be unparsed
 TEST_F(CfgHostsTest, unparse6) {
     CfgMgr::instance().setFamily(AF_INET6);
@@ -516,7 +752,7 @@ TEST_F(CfgHostsTest, get6ByAddr) {
 
         // Add host identified by DUID.
         HostPtr host = HostPtr(new Host(duids_[i]->toText(), "duid",
-                                        SubnetID(0), SubnetID(1 + i % 2),
+                                        SUBNET_ID_UNUSED, SubnetID(1 + i % 2),
                                         IOAddress("0.0.0.0")));
         host->addReservation(IPv6Resrv(IPv6Resrv::TYPE_NA,
                                        increase(IOAddress("2001:db8:2::1"),
@@ -549,7 +785,7 @@ TEST_F(CfgHostsTest, get6MultipleAddrs) {
 
         // Add host identified by DUID.
         HostPtr host = HostPtr(new Host(duids_[i]->toText(), "duid",
-                                        SubnetID(0), SubnetID(1 + i % 2),
+                                        SUBNET_ID_UNUSED, SubnetID(1 + i % 2),
                                         IOAddress("0.0.0.0")));
 
         // Generate 5 unique addresses for this host.
@@ -599,7 +835,7 @@ TEST_F(CfgHostsTest, add4AlreadyReserved) {
     // First host has a reservation for address 192.0.2.1
     HostPtr host1 = HostPtr(new Host(hwaddrs_[0]->toText(false),
                                      "hw-address",
-                                     SubnetID(1), SubnetID(0),
+                                     SubnetID(1), SubnetID(SUBNET_ID_UNUSED),
                                      IOAddress("192.0.2.1")));
     // Adding this should work.
     EXPECT_NO_THROW(cfg.add(host1));
@@ -607,7 +843,7 @@ TEST_F(CfgHostsTest, add4AlreadyReserved) {
     // The second host has a reservation for the same address.
     HostPtr host2 = HostPtr(new Host(hwaddrs_[1]->toText(false),
                                      "hw-address",
-                                     SubnetID(1), SubnetID(0),
+                                     SubnetID(1), SUBNET_ID_UNUSED,
                                      IOAddress("192.0.2.1")));
 
     // This second host has a reservation for an address that is already
@@ -622,7 +858,7 @@ TEST_F(CfgHostsTest, add6Invalid2Hosts) {
 
     // First host has a reservation for address 2001:db8::1
     HostPtr host1 = HostPtr(new Host(duids_[0]->toText(), "duid",
-                                     SubnetID(0), SubnetID(1),
+                                     SUBNET_ID_UNUSED, SubnetID(1),
                                      IOAddress("0.0.0.0")));
     host1->addReservation(IPv6Resrv(IPv6Resrv::TYPE_NA,
                                     IOAddress("2001:db8::1")));
@@ -631,7 +867,7 @@ TEST_F(CfgHostsTest, add6Invalid2Hosts) {
 
     // The second host has a reservation for the same address.
     HostPtr host2 = HostPtr(new Host(duids_[1]->toText(), "duid",
-                                     SubnetID(0), SubnetID(1),
+                                     SUBNET_ID_UNUSED, SubnetID(1),
                                      IOAddress("0.0.0.0")));
     host2->addReservation(IPv6Resrv(IPv6Resrv::TYPE_NA,
                                     IOAddress("2001:db8::1")));
@@ -641,13 +877,24 @@ TEST_F(CfgHostsTest, add6Invalid2Hosts) {
     EXPECT_THROW(cfg.add(host2), isc::dhcp::DuplicateHost);
 }
 
+// Check that no error is reported when adding a host with subnet
+// ids equal to global.
+TEST_F(CfgHostsTest, globalSubnetIDs) {
+    CfgHosts cfg;
+    ASSERT_NO_THROW(cfg.add(HostPtr(new Host(hwaddrs_[0]->toText(false),
+                                             "hw-address",
+                                             SUBNET_ID_GLOBAL, SUBNET_ID_GLOBAL,
+                                             IOAddress("10.0.0.1")))));
+}
+
+
 // Check that error is reported when trying to add a host with subnet
-// ids equal to zero.
-TEST_F(CfgHostsTest, zeroSubnetIDs) {
+// ids equal to unused.
+TEST_F(CfgHostsTest, unusedSubnetIDs) {
     CfgHosts cfg;
     ASSERT_THROW(cfg.add(HostPtr(new Host(hwaddrs_[0]->toText(false),
                                           "hw-address",
-                                          SubnetID(0), SubnetID(0),
+                                          SUBNET_ID_UNUSED, SUBNET_ID_UNUSED,
                                           IOAddress("10.0.0.1")))),
                  isc::BadValue);
 }
@@ -659,21 +906,21 @@ TEST_F(CfgHostsTest, duplicatesSubnet4HWAddr) {
     // Add a host.
     ASSERT_NO_THROW(cfg.add(HostPtr(new Host(hwaddrs_[0]->toText(false),
                                              "hw-address",
-                                             SubnetID(10), SubnetID(0),
+                                             SubnetID(10), SUBNET_ID_UNUSED,
                                              IOAddress("10.0.0.1")))));
 
     // Try to add the host with the same HW address to the same subnet. The fact
     // that the IP address is different here shouldn't really matter.
     EXPECT_THROW(cfg.add(HostPtr(new Host(hwaddrs_[0]->toText(false),
                                           "hw-address",
-                                          SubnetID(10), SubnetID(0),
+                                          SubnetID(10), SUBNET_ID_UNUSED,
                                           IOAddress("10.0.0.10")))),
                  isc::dhcp::DuplicateHost);
 
     // Now try to add it to a different subnet. It should go through.
     EXPECT_NO_THROW(cfg.add(HostPtr(new Host(hwaddrs_[0]->toText(false),
                                              "hw-address",
-                                             SubnetID(11), SubnetID(0),
+                                             SubnetID(11), SUBNET_ID_UNUSED,
                                              IOAddress("10.0.0.10")))));
 }
 
@@ -684,21 +931,21 @@ TEST_F(CfgHostsTest, duplicatesSubnet4DUID) {
     // Add a host.
     ASSERT_NO_THROW(cfg.add(HostPtr(new Host(duids_[0]->toText(),
                                              "duid",
-                                             SubnetID(10), SubnetID(0),
+                                             SubnetID(10), SUBNET_ID_UNUSED,
                                              IOAddress("10.0.0.1")))));
 
     // Try to add the host with the same DUID to the same subnet. The fact
     // that the IP address is different here shouldn't really matter.
     EXPECT_THROW(cfg.add(HostPtr(new Host(duids_[0]->toText(),
                                           "duid",
-                                          SubnetID(10), SubnetID(0),
+                                          SubnetID(10), SUBNET_ID_UNUSED,
                                           IOAddress("10.0.0.10")))),
                  isc::dhcp::DuplicateHost);
 
     // Now try to add it to a different subnet. It should go through.
     EXPECT_NO_THROW(cfg.add(HostPtr(new Host(duids_[0]->toText(),
                                              "duid",
-                                             SubnetID(11), SubnetID(0),
+                                             SubnetID(11), SUBNET_ID_UNUSED,
                                              IOAddress("10.0.0.10")))));
 }
 
@@ -709,7 +956,7 @@ TEST_F(CfgHostsTest, duplicatesSubnet6HWAddr) {
     // Add a host.
     ASSERT_NO_THROW(cfg.add(HostPtr(new Host(hwaddrs_[0]->toText(false),
                                              "hw-address",
-                                             SubnetID(0), SubnetID(1),
+                                             SUBNET_ID_UNUSED, SubnetID(1),
                                              IOAddress("0.0.0.0"),
                                              "foo.example.com"))));
 
@@ -717,7 +964,7 @@ TEST_F(CfgHostsTest, duplicatesSubnet6HWAddr) {
     // that the IP address is different here shouldn't really matter.
     EXPECT_THROW(cfg.add(HostPtr(new Host(hwaddrs_[0]->toText(false),
                                           "hw-address",
-                                          SubnetID(0), SubnetID(1),
+                                          SUBNET_ID_UNUSED, SubnetID(1),
                                           IOAddress("0.0.0.0"),
                                           "foo.example.com"))),
                  isc::dhcp::DuplicateHost);
@@ -725,7 +972,7 @@ TEST_F(CfgHostsTest, duplicatesSubnet6HWAddr) {
     // Now try to add it to a different subnet. It should go through.
     EXPECT_NO_THROW(cfg.add(HostPtr(new Host(hwaddrs_[0]->toText(false),
                                              "hw-address",
-                                             SubnetID(0), SubnetID(2),
+                                             SUBNET_ID_UNUSED, SubnetID(2),
                                              IOAddress("0.0.0.0"),
                                              "foo.example.com"))));
 }
@@ -737,7 +984,7 @@ TEST_F(CfgHostsTest, duplicatesSubnet6DUID) {
     // Add a host.
     ASSERT_NO_THROW(cfg.add(HostPtr(new Host(duids_[0]->toText(),
                                              "duid",
-                                             SubnetID(0), SubnetID(1),
+                                             SUBNET_ID_UNUSED, SubnetID(1),
                                              IOAddress("0.0.0.0"),
                                              "foo.example.com"))));
 
@@ -745,7 +992,7 @@ TEST_F(CfgHostsTest, duplicatesSubnet6DUID) {
     // that the IP address is different here shouldn't really matter.
     EXPECT_THROW(cfg.add(HostPtr(new Host(duids_[0]->toText(),
                                           "duid",
-                                          SubnetID(0), SubnetID(1),
+                                          SUBNET_ID_UNUSED, SubnetID(1),
                                           IOAddress("0.0.0.0"),
                                           "foo.example.com"))),
                  isc::dhcp::DuplicateHost);
@@ -753,7 +1000,7 @@ TEST_F(CfgHostsTest, duplicatesSubnet6DUID) {
     // Now try to add it to a different subnet. It should go through.
     EXPECT_NO_THROW(cfg.add(HostPtr(new Host(duids_[0]->toText(),
                                              "duid",
-                                             SubnetID(0), SubnetID(2),
+                                             SUBNET_ID_UNUSED, SubnetID(2),
                                              IOAddress("0.0.0.0"),
                                              "foo.example.com"))));
 }
