@@ -1,4 +1,4 @@
-// Copyright (C) 2014-2017 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2014-2019 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -6,13 +6,23 @@
 
 #include <config.h>
 #include <dhcp/classify.h>
+#include <dhcp/libdhcp++.h>
+#include <dhcp/option_custom.h>
+#include <dhcp/option_definition.h>
+#include <dhcp/option_space.h>
+#include <dhcp/option_string.h>
 #include <dhcp/tests/iface_mgr_test_config.h>
+#include <dhcpsrv/parsers/dhcp_parsers.h>
 #include <dhcpsrv/shared_network.h>
+#include <dhcpsrv/cfg_shared_networks.h>
 #include <dhcpsrv/cfg_subnets4.h>
+#include <dhcpsrv/shared_network.h>
 #include <dhcpsrv/subnet.h>
 #include <dhcpsrv/subnet_id.h>
 #include <dhcpsrv/subnet_selector.h>
 #include <testutils/test_to_element.h>
+#include <util/doubles.h>
+
 #include <gtest/gtest.h>
 #include <vector>
 
@@ -24,6 +34,38 @@ using namespace isc::test;
 
 namespace {
 
+/// @brief Verifies that a set of subnets contains a given a subnet
+///
+/// @param cfg_subnets set of sunbets in which to look
+/// @param prefix prefix of the target subnet
+/// @param exp_subnet_id expected id of the target subnet
+/// @param exp_valid expected valid lifetime of the subnet
+/// @param exp_network  pointer to the subnet's shared-network (if one)
+void checkMergedSubnet(CfgSubnets4& cfg_subnets,
+                       const std::string& prefix,
+                       const SubnetID exp_subnet_id,
+                       int exp_valid,
+                       SharedNetwork4Ptr exp_network) {
+    // Look for the network by prefix.
+    auto subnet = cfg_subnets.getByPrefix(prefix);
+    ASSERT_TRUE(subnet) << "subnet: " << prefix << " not found";
+
+    // Make sure we have the one we expect.
+    ASSERT_EQ(exp_subnet_id, subnet->getID()) << "subnet ID is wrong";
+    ASSERT_EQ(exp_valid, subnet->getValid()) << "subnet valid time is wrong";
+
+    SharedNetwork4Ptr shared_network;
+    subnet->getSharedNetwork(shared_network);
+    if (exp_network) {
+        ASSERT_TRUE(shared_network)
+            << " expected network: " << exp_network->getName() << " not found";
+        ASSERT_TRUE(shared_network == exp_network) << " networks do no match";
+    } else {
+        ASSERT_FALSE(shared_network) << " unexpected network assignment: "
+            << shared_network->getName();
+    }
+}
+
 // This test verifies that specific subnet can be retrieved by specifying
 // subnet identifier or subnet prefix.
 TEST(CfgSubnets4Test, getSpecificSubnet) {
@@ -101,6 +143,241 @@ TEST(CfgSubnets4Test, deleteSubnet) {
     ASSERT_NO_THROW(cfg.del(subnet2));
     ASSERT_EQ(2, cfg.getAll()->size());
     EXPECT_FALSE(cfg.getByPrefix("192.0.3.0/26"));
+
+    // Remove another subnet by ID.
+    ASSERT_NO_THROW(cfg.del(subnet1->getID()));
+    ASSERT_EQ(1, cfg.getAll()->size());
+    EXPECT_FALSE(cfg.getByPrefix("192.0.2.0/26"));
+}
+
+// This test verifies that replace a subnet works as expected.
+TEST(CfgSubnets4Test, replaceSubnet) {
+    CfgSubnets4 cfg;
+
+    // Create 3 subnets.
+    Subnet4Ptr subnet1(new Subnet4(IOAddress("192.0.1.0"),
+                                   26, 1, 2, 100, SubnetID(10)));
+    Subnet4Ptr subnet2(new Subnet4(IOAddress("192.0.2.0"),
+                                   26, 1, 2, 100, SubnetID(2)));
+    Subnet4Ptr subnet3(new Subnet4(IOAddress("192.0.3.0"),
+                                   26, 1, 2, 100, SubnetID(13)));
+
+    ASSERT_NO_THROW(cfg.add(subnet1));
+    ASSERT_NO_THROW(cfg.add(subnet2));
+    ASSERT_NO_THROW(cfg.add(subnet3));
+
+    // There should be three subnets.
+    ASSERT_EQ(3, cfg.getAll()->size());
+    // We're going to replace  the subnet #2. Let's make sure it exists before
+    // we replace it.
+    ASSERT_TRUE(cfg.getByPrefix("192.0.3.0/26"));
+
+    // Replace the subnet and make sure it was updated.
+    Subnet4Ptr subnet(new Subnet4(IOAddress("192.0.2.0"),
+                                  26, 10, 20, 1000,  SubnetID(2)));
+    Subnet4Ptr replaced = cfg.replace(subnet);
+    ASSERT_TRUE(replaced);
+    EXPECT_TRUE(replaced == subnet2);
+    ASSERT_EQ(3, cfg.getAll()->size());
+    Subnet4Ptr returned = cfg.getAll()->at(1);
+    ASSERT_TRUE(returned);
+    EXPECT_TRUE(returned == subnet);
+
+    // Restore.
+    replaced = cfg.replace(replaced);
+    ASSERT_TRUE(replaced);
+    EXPECT_TRUE(replaced == subnet);
+    ASSERT_EQ(3, cfg.getAll()->size());
+    returned = cfg.getAll()->at(1);
+    ASSERT_TRUE(returned);
+    EXPECT_TRUE(returned == subnet2);
+
+    // Prefix conflict returns null.
+    subnet.reset(new Subnet4(IOAddress("192.0.3.0"),
+                             26, 10, 20, 1000,  SubnetID(2)));
+    replaced = cfg.replace(subnet);
+    EXPECT_FALSE(replaced);
+    returned = cfg.getAll()->at(1);
+    ASSERT_TRUE(returned);
+    EXPECT_TRUE(returned == subnet2);
+
+    // Changing prefix works even it is highly not recommended.
+    subnet.reset(new Subnet4(IOAddress("192.0.10.0"),
+                             26, 10, 20, 1000,  SubnetID(2)));
+    replaced = cfg.replace(subnet);
+    ASSERT_TRUE(replaced);
+    EXPECT_TRUE(replaced == subnet2);
+    returned = cfg.getAll()->at(1);
+    ASSERT_TRUE(returned);
+    EXPECT_TRUE(returned == subnet);
+}
+
+// This test verifies that subnets configuration is properly merged.
+TEST(CfgSubnets4Test, mergeSubnets) {
+    // Create custom options dictionary for testing merge. We're keeping it
+    // simple because they are more rigorous tests elsewhere.
+    CfgOptionDefPtr cfg_def(new CfgOptionDef());
+    cfg_def->add((OptionDefinitionPtr(new OptionDefinition("one", 1, "string"))), "isc");
+
+    Subnet4Ptr subnet1(new Subnet4(IOAddress("192.0.1.0"),
+                                   26, 1, 2, 100, SubnetID(1)));
+    Subnet4Ptr subnet2(new Subnet4(IOAddress("192.0.2.0"),
+                                   26, 1, 2, 100, SubnetID(2)));
+    Subnet4Ptr subnet3(new Subnet4(IOAddress("192.0.3.0"),
+                                   26, 1, 2, 100, SubnetID(3)));
+    Subnet4Ptr subnet4(new Subnet4(IOAddress("192.0.4.0"),
+                                   26, 1, 2, 100, SubnetID(4)));
+
+    // Create the "existing" list of shared networks
+    CfgSharedNetworks4Ptr networks(new CfgSharedNetworks4());
+    SharedNetwork4Ptr shared_network1(new SharedNetwork4("shared-network1"));
+    networks->add(shared_network1);
+    SharedNetwork4Ptr shared_network2(new SharedNetwork4("shared-network2"));
+    networks->add(shared_network2);
+
+    // Empty network pointer.
+    SharedNetwork4Ptr no_network;
+
+    // Add Subnets 1, 2 and 4 to shared networks.
+    ASSERT_NO_THROW(shared_network1->add(subnet1));
+    ASSERT_NO_THROW(shared_network2->add(subnet2));
+    ASSERT_NO_THROW(shared_network2->add(subnet4));
+
+    // Create our "existing" configured subnets.
+    CfgSubnets4 cfg_to;
+    ASSERT_NO_THROW(cfg_to.add(subnet1));
+    ASSERT_NO_THROW(cfg_to.add(subnet2));
+    ASSERT_NO_THROW(cfg_to.add(subnet3));
+    ASSERT_NO_THROW(cfg_to.add(subnet4));
+
+    // Merge in an "empty" config. Should have the original config,
+    // still intact.
+    CfgSubnets4 cfg_from;
+    ASSERT_NO_THROW(cfg_to.merge(cfg_def, networks, cfg_from));
+
+    // We should have all four subnets, with no changes.
+    ASSERT_EQ(4, cfg_to.getAll()->size());
+
+    // Should be no changes to the configuration.
+    ASSERT_NO_FATAL_FAILURE(checkMergedSubnet(cfg_to, "192.0.1.0/26",
+                                              SubnetID(1), 100, shared_network1));
+    ASSERT_NO_FATAL_FAILURE(checkMergedSubnet(cfg_to, "192.0.2.0/26",
+                                              SubnetID(2), 100, shared_network2));
+    ASSERT_NO_FATAL_FAILURE(checkMergedSubnet(cfg_to, "192.0.3.0/26",
+                                              SubnetID(3), 100, no_network));
+    ASSERT_NO_FATAL_FAILURE(checkMergedSubnet(cfg_to, "192.0.4.0/26",
+                                              SubnetID(4), 100, shared_network2));
+
+    // Fill cfg_from configuration with subnets.
+    // subnet 1b updates subnet 1 but leaves it in network 1 with the same ID.
+    Subnet4Ptr subnet1b(new Subnet4(IOAddress("192.0.10.0"),
+                                   26, 2, 3, 400, SubnetID(1)));
+    subnet1b->setSharedNetworkName("shared-network1");
+
+    // Add generic option 1 to subnet 1b.
+    std::string value("Yay!");
+    OptionPtr option(new Option(Option::V4, 1));
+    option->setData(value.begin(), value.end());
+    ASSERT_NO_THROW(subnet1b->getCfgOption()->add(option, false, "isc"));
+
+    // subnet 3b updates subnet 3 with different ID and removes it
+    // from network 2
+    Subnet4Ptr subnet3b(new Subnet4(IOAddress("192.0.3.0"),
+                                   26, 3, 4, 500, SubnetID(30)));
+
+    // Now Add generic option 1 to subnet 3b.
+    value = "Team!";
+    option.reset(new Option(Option::V4, 1));
+    option->setData(value.begin(), value.end());
+    ASSERT_NO_THROW(subnet3b->getCfgOption()->add(option, false, "isc"));
+
+    // subnet 4b updates subnet 4 and moves it from network2 to network 1
+    Subnet4Ptr subnet4b(new Subnet4(IOAddress("192.0.4.0"),
+                                   26, 3, 4, 500, SubnetID(4)));
+    subnet4b->setSharedNetworkName("shared-network1");
+
+    // subnet 5 is new and belongs to network 2
+    // Has two pools both with an option 1
+    Subnet4Ptr subnet5(new Subnet4(IOAddress("192.0.5.0"),
+                                   26, 1, 2, 300, SubnetID(5)));
+    subnet5->setSharedNetworkName("shared-network2");
+
+    // Add pool 1
+    Pool4Ptr pool(new Pool4(IOAddress("192.0.5.10"), IOAddress("192.0.5.20")));
+    value = "POOLS";
+    option.reset(new Option(Option::V4, 1));
+    option->setData(value.begin(), value.end());
+    ASSERT_NO_THROW(pool->getCfgOption()->add(option, false, "isc"));
+    subnet5->addPool(pool);
+
+    // Add pool 2
+    pool.reset(new Pool4(IOAddress("192.0.5.30"), IOAddress("192.0.5.40")));
+    value ="RULE!";
+    option.reset(new Option(Option::V4, 1));
+    option->setData(value.begin(), value.end());
+    ASSERT_NO_THROW(pool->getCfgOption()->add(option, false, "isc"));
+    subnet5->addPool(pool);
+
+    // Add subnets to the merge from config.
+    ASSERT_NO_THROW(cfg_from.add(subnet1b));
+    ASSERT_NO_THROW(cfg_from.add(subnet3b));
+    ASSERT_NO_THROW(cfg_from.add(subnet4b));
+    ASSERT_NO_THROW(cfg_from.add(subnet5));
+
+    // Merge again.
+    ASSERT_NO_THROW(cfg_to.merge(cfg_def, networks, cfg_from));
+    ASSERT_EQ(5, cfg_to.getAll()->size());
+
+    // The subnet1 should be replaced by subnet1b.
+    ASSERT_NO_FATAL_FAILURE(checkMergedSubnet(cfg_to, "192.0.10.0/26",
+                                              SubnetID(1), 400, shared_network1));
+
+    // Let's verify that our option is there and populated correctly.
+    auto subnet = cfg_to.getByPrefix("192.0.10.0/26");
+    auto desc = subnet->getCfgOption()->get("isc", 1);
+    ASSERT_TRUE(desc.option_);
+    OptionStringPtr opstr = boost::dynamic_pointer_cast<OptionString>(desc.option_);
+    ASSERT_TRUE(opstr);
+    EXPECT_EQ("Yay!", opstr->getValue());
+
+    // The subnet2 should not be affected because it was not present.
+    ASSERT_NO_FATAL_FAILURE(checkMergedSubnet(cfg_to, "192.0.2.0/26",
+                                              SubnetID(2), 100, shared_network2));
+
+    // subnet3 should be replaced by subnet3b and no longer assigned to a network.
+    ASSERT_NO_FATAL_FAILURE(checkMergedSubnet(cfg_to, "192.0.3.0/26",
+                                              SubnetID(30), 500, no_network));
+    // Let's verify that our option is there and populated correctly.
+    subnet = cfg_to.getByPrefix("192.0.3.0/26");
+    desc = subnet->getCfgOption()->get("isc", 1);
+    ASSERT_TRUE(desc.option_);
+    opstr = boost::dynamic_pointer_cast<OptionString>(desc.option_);
+    ASSERT_TRUE(opstr);
+    EXPECT_EQ("Team!", opstr->getValue());
+
+    // subnet4 should be replaced by subnet4b and moved to network1.
+    ASSERT_NO_FATAL_FAILURE(checkMergedSubnet(cfg_to, "192.0.4.0/26",
+                                              SubnetID(4), 500, shared_network1));
+
+    // subnet5 should have been added to configuration.
+    ASSERT_NO_FATAL_FAILURE(checkMergedSubnet(cfg_to, "192.0.5.0/26",
+                                              SubnetID(5), 300, shared_network2));
+
+    // Let's verify that both pools have the proper options.
+    subnet = cfg_to.getByPrefix("192.0.5.0/26");
+    const PoolPtr merged_pool = subnet->getPool(Lease::TYPE_V4, IOAddress("192.0.5.10"));
+    ASSERT_TRUE(merged_pool);
+    desc = merged_pool->getCfgOption()->get("isc", 1);
+    opstr = boost::dynamic_pointer_cast<OptionString>(desc.option_);
+    ASSERT_TRUE(opstr);
+    EXPECT_EQ("POOLS", opstr->getValue());
+
+    const PoolPtr merged_pool2 = subnet->getPool(Lease::TYPE_V4, IOAddress("192.0.5.30"));
+    ASSERT_TRUE(merged_pool2);
+    desc = merged_pool2->getCfgOption()->get("isc", 1);
+    opstr = boost::dynamic_pointer_cast<OptionString>(desc.option_);
+    ASSERT_TRUE(opstr);
+    EXPECT_EQ("RULE!", opstr->getValue());
 }
 
 // This test verifies that it is possible to retrieve a subnet using an
@@ -410,7 +687,7 @@ TEST(CfgSubnets4Test, selectSubnetByOptionSelect) {
 
     // Over relay-info too
     selector.giaddr_ = IOAddress("10.0.0.1");
-    subnet2->setRelayInfo(IOAddress("10.0.0.1"));
+    subnet2->addRelayAddress(IOAddress("10.0.0.1"));
     EXPECT_EQ(subnet3, cfg.selectSubnet(selector));
     selector.option_select_ = IOAddress("0.0.0.0");
     EXPECT_EQ(subnet2, cfg.selectSubnet(selector));
@@ -446,9 +723,9 @@ TEST(CfgSubnets4Test, selectSubnetByRelayAddress) {
     EXPECT_FALSE(cfg.selectSubnet(selector));
 
     // Now specify relay info
-    subnet1->setRelayInfo(IOAddress("10.0.0.1"));
-    subnet2->setRelayInfo(IOAddress("10.0.0.2"));
-    subnet3->setRelayInfo(IOAddress("10.0.0.3"));
+    subnet1->addRelayAddress(IOAddress("10.0.0.1"));
+    subnet2->addRelayAddress(IOAddress("10.0.0.2"));
+    subnet3->addRelayAddress(IOAddress("10.0.0.3"));
 
     // And try again. This time relay-info is there and should match.
     selector.giaddr_ = IOAddress("10.0.0.1");
@@ -481,9 +758,9 @@ TEST(CfgSubnets4Test, selectSharedNetworkByRelayAddressNetworkLevel) {
 
     // Now specify relay info. Note that for the second subnet we specify
     // relay info on the network level.
-    subnet1->setRelayInfo(IOAddress("10.0.0.1"));
-    network->setRelayInfo(IOAddress("10.0.0.2"));
-    subnet3->setRelayInfo(IOAddress("10.0.0.3"));
+    subnet1->addRelayAddress(IOAddress("10.0.0.1"));
+    network->addRelayAddress(IOAddress("10.0.0.2"));
+    subnet3->addRelayAddress(IOAddress("10.0.0.3"));
 
     // And try again. This time relay-info is there and should match.
     selector.giaddr_ = IOAddress("10.0.0.1");
@@ -520,9 +797,9 @@ TEST(CfgSubnets4Test, selectSharedNetworkByRelayAddressSubnetLevel) {
 
     // Now specify relay info. Note that for the second subnet we specify
     // relay info on the network level.
-    subnet1->setRelayInfo(IOAddress("10.0.0.1"));
-    subnet2->setRelayInfo(IOAddress("10.0.0.2"));
-    subnet3->setRelayInfo(IOAddress("10.0.0.3"));
+    subnet1->addRelayAddress(IOAddress("10.0.0.1"));
+    subnet2->addRelayAddress(IOAddress("10.0.0.2"));
+    subnet3->addRelayAddress(IOAddress("10.0.0.3"));
 
     // And try again. This time relay-info is there and should match.
     selector.giaddr_ = IOAddress("10.0.0.1");
@@ -731,10 +1008,32 @@ TEST(CfgSubnets4Test, unparseSubnet) {
     Subnet4Ptr subnet1(new Subnet4(IOAddress("192.0.2.0"), 26, 1, 2, 3, 123));
     Subnet4Ptr subnet2(new Subnet4(IOAddress("192.0.2.64"), 26, 1, 2, 3, 124));
     Subnet4Ptr subnet3(new Subnet4(IOAddress("192.0.2.128"), 26, 1, 2, 3, 125));
+
     subnet1->allowClientClass("foo");
+
+    subnet1->setT1Percent(0.45);
+    subnet1->setT2Percent(0.70);
+
     subnet2->setIface("lo");
-    subnet2->setRelayInfo(IOAddress("10.0.0.1"));
+    subnet2->addRelayAddress(IOAddress("10.0.0.1"));
+
     subnet3->setIface("eth1");
+    subnet3->requireClientClass("foo");
+    subnet3->requireClientClass("bar");
+    subnet3->setCalculateTeeTimes(true);
+    subnet3->setT1Percent(0.50);
+    subnet3->setT2Percent(0.65);
+    subnet3->setHostReservationMode(Network::HR_ALL);
+    subnet3->setAuthoritative(false);
+    subnet3->setMatchClientId(true);
+    subnet3->setSiaddr(IOAddress("192.0.2.2"));
+    subnet3->setSname("frog");
+    subnet3->setFilename("/dev/null");
+
+    data::ElementPtr ctx1 = data::Element::fromJSON("{ \"comment\": \"foo\" }");
+    subnet1->setContext(ctx1);
+    data::ElementPtr ctx2 = data::Element::createMap();
+    subnet2->setContext(ctx2);
 
     cfg.add(subnet1);
     cfg.add(subnet2);
@@ -743,53 +1042,58 @@ TEST(CfgSubnets4Test, unparseSubnet) {
     // Unparse
     std::string expected = "[\n"
         "{\n"
+        "    \"comment\": \"foo\",\n"
         "    \"id\": 123,\n"
         "    \"subnet\": \"192.0.2.0/26\",\n"
-        "    \"relay\": { \"ip-address\": \"0.0.0.0\" },\n"
-        "    \"match-client-id\": true,\n"
-        "    \"next-server\": \"0.0.0.0\",\n"
+        "    \"t1-percent\": 0.45,"
+        "    \"t2-percent\": 0.7,"
         "    \"renew-timer\": 1,\n"
         "    \"rebind-timer\": 2,\n"
+        "    \"relay\": { \"ip-addresses\": [ ] },\n"
         "    \"valid-lifetime\": 3,\n"
         "    \"client-class\": \"foo\",\n"
         "    \"4o6-interface\": \"\",\n"
         "    \"4o6-interface-id\": \"\",\n"
         "    \"4o6-subnet\": \"\",\n"
-        "    \"reservation-mode\": \"all\",\n"
         "    \"option-data\": [ ],\n"
         "    \"pools\": [ ]\n"
         "},{\n"
         "    \"id\": 124,\n"
         "    \"subnet\": \"192.0.2.64/26\",\n"
-        "    \"relay\": { \"ip-address\": \"10.0.0.1\" },\n"
         "    \"interface\": \"lo\",\n"
-        "    \"match-client-id\": true,\n"
-        "    \"next-server\": \"0.0.0.0\",\n"
         "    \"renew-timer\": 1,\n"
         "    \"rebind-timer\": 2,\n"
+        "    \"relay\": { \"ip-addresses\": [ \"10.0.0.1\" ] },\n"
         "    \"valid-lifetime\": 3,\n"
         "    \"4o6-interface\": \"\",\n"
         "    \"4o6-interface-id\": \"\",\n"
         "    \"4o6-subnet\": \"\",\n"
-        "    \"reservation-mode\": \"all\",\n"
+        "    \"user-context\": {},\n"
         "    \"option-data\": [ ],\n"
         "    \"pools\": [ ]\n"
         "},{\n"
         "    \"id\": 125,\n"
         "    \"subnet\": \"192.0.2.128/26\",\n"
-        "    \"relay\": { \"ip-address\": \"0.0.0.0\" },\n"
         "    \"interface\": \"eth1\",\n"
         "    \"match-client-id\": true,\n"
-        "    \"next-server\": \"0.0.0.0\",\n"
+        "    \"next-server\": \"192.0.2.2\",\n"
+        "    \"server-hostname\": \"frog\",\n"
+        "    \"boot-file-name\": \"/dev/null\",\n"
         "    \"renew-timer\": 1,\n"
         "    \"rebind-timer\": 2,\n"
+        "    \"relay\": { \"ip-addresses\": [ ] },\n"
         "    \"valid-lifetime\": 3,\n"
         "    \"4o6-interface\": \"\",\n"
         "    \"4o6-interface-id\": \"\",\n"
         "    \"4o6-subnet\": \"\",\n"
+        "    \"authoritative\": false,\n"
         "    \"reservation-mode\": \"all\",\n"
         "    \"option-data\": [ ],\n"
-        "    \"pools\": [ ]\n"
+        "    \"pools\": [ ]\n,"
+        "    \"require-client-classes\": [ \"foo\", \"bar\" ],\n"
+        "    \"calculate-tee-times\": true,\n"
+        "    \"t1-percent\": 0.50,\n"
+        "    \"t2-percent\": 0.65\n"
         "} ]\n";
     runToElementTest<CfgSubnets4>(expected, cfg);
 }
@@ -802,6 +1106,14 @@ TEST(CfgSubnets4Test, unparsePool) {
     Subnet4Ptr subnet(new Subnet4(IOAddress("192.0.2.0"), 24, 1, 2, 3, 123));
     Pool4Ptr pool1(new Pool4(IOAddress("192.0.2.1"), IOAddress("192.0.2.10")));
     Pool4Ptr pool2(new Pool4(IOAddress("192.0.2.64"), 26));
+    pool2->allowClientClass("bar");
+
+    std::string json1 = "{ \"comment\": \"foo\", \"version\": 1 }";
+    data::ElementPtr ctx1 = data::Element::fromJSON(json1);
+    pool1->setContext(ctx1);
+    data::ElementPtr ctx2 = data::Element::fromJSON("{ \"foo\": \"bar\" }");
+    pool2->setContext(ctx2);
+    pool2->requireClientClass("foo");
 
     subnet->addPool(pool1);
     subnet->addPool(pool2);
@@ -812,24 +1124,26 @@ TEST(CfgSubnets4Test, unparsePool) {
         "{\n"
         "    \"id\": 123,\n"
         "    \"subnet\": \"192.0.2.0/24\",\n"
-        "    \"relay\": { \"ip-address\": \"0.0.0.0\" },\n"
-        "    \"match-client-id\": true,\n"
-        "    \"next-server\": \"0.0.0.0\",\n"
         "    \"renew-timer\": 1,\n"
         "    \"rebind-timer\": 2,\n"
+        "    \"relay\": { \"ip-addresses\": [ ] },\n"
         "    \"valid-lifetime\": 3,\n"
         "    \"4o6-interface\": \"\",\n"
         "    \"4o6-interface-id\": \"\",\n"
         "    \"4o6-subnet\": \"\",\n"
-        "    \"reservation-mode\": \"all\",\n"
         "    \"option-data\": [],\n"
         "    \"pools\": [\n"
         "        {\n"
+        "            \"comment\": \"foo\",\n"
         "            \"option-data\": [ ],\n"
-        "            \"pool\": \"192.0.2.1-192.0.2.10\"\n"
+        "            \"pool\": \"192.0.2.1-192.0.2.10\",\n"
+        "            \"user-context\": { \"version\": 1 }\n"
         "        },{\n"
         "            \"option-data\": [ ],\n"
-        "            \"pool\": \"192.0.2.64/26\"\n"
+        "            \"pool\": \"192.0.2.64/26\",\n"
+        "            \"user-context\": { \"foo\": \"bar\" },\n"
+        "            \"client-class\": \"bar\",\n"
+        "            \"require-client-classes\": [ \"foo\" ]\n"
         "        }\n"
         "    ]\n"
         "} ]\n";
@@ -856,4 +1170,129 @@ TEST(CfgSubnets4Test, getSubnet) {
     EXPECT_EQ(Subnet4Ptr(), cfg.getSubnet(400)); // no such subnet
 }
 
+// This test verifies that hasSubnetWithServerId returns correct value.
+TEST(CfgSubnets4Test, hasSubnetWithServerId) {
+    CfgSubnets4 cfg;
+
+    // Initially, there is no server identifier option present.
+    EXPECT_FALSE(cfg.hasSubnetWithServerId(IOAddress("1.2.3.4")));
+
+    OptionDefinitionPtr def = LibDHCP::getOptionDef(DHCP4_OPTION_SPACE,
+                                                    DHO_DHCP_SERVER_IDENTIFIER);
+    OptionCustomPtr opt_server_id(new OptionCustom(*def, Option::V4));
+    opt_server_id->writeAddress(IOAddress("1.2.3.4"));
+    Subnet4Ptr subnet(new Subnet4(IOAddress("192.0.2.0"), 26, 1, 2, 3, 100));
+    subnet->getCfgOption()->add(opt_server_id, false, DHCP4_OPTION_SPACE);
+    cfg.add(subnet);
+
+    EXPECT_TRUE(cfg.hasSubnetWithServerId(IOAddress("1.2.3.4")));
+    EXPECT_FALSE(cfg.hasSubnetWithServerId(IOAddress("2.3.4.5")));
+}
+
+// This test verifies the Subnet4 parser's validation logic for
+// t1-percent and t2-percent parameters.
+TEST(CfgSubnets4Test, teeTimePercentValidation) {
+
+    // Describes a single test scenario.
+    struct Scenario {
+        std::string label;         // label used for logging test failures
+        bool calculate_tee_times;  // value of calculate-tee-times parameter
+        double t1_percent;         // value of t1-percent parameter
+        double t2_percent;         // value of t2-percent parameter
+        std::string error_message; // expected error message is parsing should fail
+    };
+
+    // Test Scenarios.
+    std::vector<Scenario> tests = {
+        {"off and valid", false, .5, .95, ""},
+        {"on and valid", true, .5, .95, ""},
+        {"t2_negative", true, .5, -.95,
+         "subnet configuration failed: t2-percent:"
+         "  -0.95 is invalid, it must be greater than 0.0 and less than 1.0"
+        },
+        {"t2_too_big", true, .5, 1.95,
+         "subnet configuration failed: t2-percent:"
+         "  1.95 is invalid, it must be greater than 0.0 and less than 1.0"
+        },
+        {"t1_negative", true, -.5, .95,
+         "subnet configuration failed: t1-percent:"
+         "  -0.5 is invalid it must be greater than 0.0 and less than 1.0"
+        },
+        {"t1_too_big", true, 1.5, .95,
+         "subnet configuration failed: t1-percent:"
+         "  1.5 is invalid it must be greater than 0.0 and less than 1.0"
+        },
+        {"t1_bigger_than_t2", true, .85, .45,
+         "subnet configuration failed: t1-percent:"
+         "  0.85 is invalid, it must be less than t2-percent: 0.45"
+        }
+    };
+
+    // First we create a set of elements that provides all
+    // required for a Subnet4.
+    std::string json =
+        "        {"
+        "            \"id\": 1,\n"
+        "            \"subnet\": \"10.1.2.0/24\", \n"
+        "            \"interface\": \"\", \n"
+        "            \"renew-timer\": 100, \n"
+        "            \"rebind-timer\": 200, \n"
+        "            \"valid-lifetime\": 300, \n"
+        "            \"match-client-id\": false, \n"
+        "            \"authoritative\": false, \n"
+        "            \"next-server\": \"\", \n"
+        "            \"server-hostname\": \"\", \n"
+        "            \"boot-file-name\": \"\", \n"
+        "            \"client-class\": \"\", \n"
+        "            \"require-client-classes\": [] \n,"
+        "            \"reservation-mode\": \"all\", \n"
+        "            \"4o6-interface\": \"\", \n"
+        "            \"4o6-interface-id\": \"\", \n"
+        "            \"4o6-subnet\": \"\", \n"
+        "            \"dhcp4o6-port\": 0, \n"
+        "            \"decline-probation-period\": 86400 \n"
+        "        }";
+
+
+    data::ElementPtr elems;
+    ASSERT_NO_THROW(elems = data::Element::fromJSON(json))
+                    << "invalid JSON:" << json << "\n test is broken";
+
+    // Iterate over the test scenarios, verifying each prescribed
+    // outcome.
+    for (auto test = tests.begin(); test != tests.end(); ++test) {
+        {
+            SCOPED_TRACE("test: " + (*test).label);
+
+            // Set this scenario's configuration parameters
+            elems->set("calculate-tee-times", data::Element::create((*test).calculate_tee_times));
+            elems->set("t1-percent", data::Element::create((*test).t1_percent));
+            elems->set("t2-percent", data::Element::create((*test).t2_percent));
+
+            Subnet4Ptr subnet;
+            try {
+                // Attempt to parse the configuration.
+                Subnet4ConfigParser parser;
+                subnet = parser.parse(elems);
+            } catch (const std::exception& ex) {
+                if (!(*test).error_message.empty()) {
+                    // We expected a failure, did we fail the correct way?
+                    EXPECT_EQ((*test).error_message, ex.what());
+                } else {
+                    // Should not have failed.
+                    ADD_FAILURE() << "Scenario should not have failed: " << ex.what();
+                }
+
+                // Either way we're done with this scenario.
+                continue;
+            }
+
+            // We parsed correctly, make sure the values are right.
+            EXPECT_EQ((*test).calculate_tee_times, subnet->getCalculateTeeTimes());
+            EXPECT_TRUE(util::areDoublesEquivalent((*test).t1_percent, subnet->getT1Percent()));
+            EXPECT_TRUE(util::areDoublesEquivalent((*test).t2_percent, subnet->getT2Percent()));
+        }
+    }
+}
+
 } // end of anonymous namespace
