@@ -1,4 +1,4 @@
-// Copyright (C) 2012-2017 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2012-2019 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -13,6 +13,7 @@
 #include <dhcpsrv/lease_mgr_factory.h>
 #include <dhcpsrv/subnet_id.h>
 #include <dhcpsrv/parsers/dhcp_parsers.h>
+#include <process/logging_info.h>
 #include <stats/stats_mgr.h>
 
 #include <boost/scoped_ptr.hpp>
@@ -31,6 +32,7 @@ using namespace isc::dhcp;
 using namespace isc::dhcp::test;
 using namespace isc::util;
 using namespace isc::stats;
+using namespace isc::process;
 using namespace isc;
 
 // don't import the entire boost namespace.  It will unexpectedly hide uint8_t
@@ -260,6 +262,7 @@ class CfgMgrTest : public ::testing::Test {
 public:
     CfgMgrTest() {
         // make sure we start with a clean configuration
+        original_datadir_ = CfgMgr::instance().getDataDir();
         clear();
     }
 
@@ -279,8 +282,8 @@ public:
     }
 
     void clear() {
-        CfgMgr::instance().setVerbose(false);
         CfgMgr::instance().setFamily(AF_INET);
+        CfgMgr::instance().setDataDir(original_datadir_);
         CfgMgr::instance().clear();
         LeaseMgrFactory::destroy();
     }
@@ -303,6 +306,10 @@ public:
 
     /// used in client classification (or just empty container for other tests)
     isc::dhcp::ClientClasses classify_;
+
+private:
+    /// to restore it in destructor.
+    string original_datadir_;
 };
 
 // Checks that there is a configuration structure available and that
@@ -318,6 +325,64 @@ TEST_F(CfgMgrTest, configuration) {
     EXPECT_TRUE(configuration->getLoggingInfo().empty());
 }
 
+// This test checks the data directory handling.
+TEST_F(CfgMgrTest, dataDir) {
+    // It is only in DHCPv6 systax so switch to IPv6.
+    CfgMgr::instance().setFamily(AF_INET6);
+
+    // Default.
+    EXPECT_TRUE(CfgMgr::instance().getDataDir().unspecified());
+    ConstElementPtr json = CfgMgr::instance().getCurrentCfg()->toElement();
+    ASSERT_TRUE(json);
+    ASSERT_EQ(Element::map, json->getType());
+    ConstElementPtr dhcp = json->get("Dhcp6");
+    ASSERT_TRUE(dhcp);
+    ASSERT_EQ(Element::map, dhcp->getType());
+    ConstElementPtr datadir = dhcp->get("data-directory");
+    EXPECT_FALSE(datadir);
+
+    // Set but not specified.
+    CfgMgr::instance().setDataDir("/tmp");
+    EXPECT_TRUE(CfgMgr::instance().getDataDir().unspecified());
+    EXPECT_EQ("/tmp", string(CfgMgr::instance().getDataDir()));
+    json = CfgMgr::instance().getCurrentCfg()->toElement();
+    ASSERT_TRUE(json);
+    ASSERT_EQ(Element::map, json->getType());
+    dhcp = json->get("Dhcp6");
+    ASSERT_TRUE(dhcp);
+    ASSERT_EQ(Element::map, dhcp->getType());
+    datadir = dhcp->get("data-directory");
+    EXPECT_FALSE(datadir);
+
+    // Set and specified.
+    CfgMgr::instance().setDataDir("/tmp", false);
+    EXPECT_FALSE(CfgMgr::instance().getDataDir().unspecified());
+    EXPECT_EQ("/tmp", string(CfgMgr::instance().getDataDir()));
+    json = CfgMgr::instance().getCurrentCfg()->toElement();
+    ASSERT_TRUE(json);
+    ASSERT_EQ(Element::map, json->getType());
+    dhcp = json->get("Dhcp6");
+    ASSERT_TRUE(dhcp);
+    ASSERT_EQ(Element::map, dhcp->getType());
+    datadir = dhcp->get("data-directory");
+    ASSERT_TRUE(datadir);
+    ASSERT_EQ(Element::string, datadir->getType());
+    EXPECT_EQ("/tmp", datadir->stringValue());
+
+    // Still IPv6 only.
+    CfgMgr::instance().setFamily(AF_INET);
+    EXPECT_FALSE(CfgMgr::instance().getDataDir().unspecified());
+    EXPECT_EQ("/tmp", string(CfgMgr::instance().getDataDir()));
+    json = CfgMgr::instance().getCurrentCfg()->toElement();
+    ASSERT_TRUE(json);
+    ASSERT_EQ(Element::map, json->getType());
+    dhcp = json->get("Dhcp4");
+    ASSERT_TRUE(dhcp);
+    ASSERT_EQ(Element::map, dhcp->getType());
+    datadir = dhcp->get("data-directory");
+    EXPECT_FALSE(datadir);
+}
+
 // This test checks the D2ClientMgr wrapper methods.
 TEST_F(CfgMgrTest, d2ClientConfig) {
     // After CfgMgr construction, D2ClientMgr member should be initialized
@@ -341,8 +406,8 @@ TEST_F(CfgMgrTest, d2ClientConfig) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  true, true, true, D2ClientConfig::RCM_ALWAYS,
-                                  "pre-fix", "suf-fix")));
+                                  true, true, D2ClientConfig::RCM_ALWAYS,
+                                  "pre-fix", "suf-fix", "[^A-z]", "*")));
 
     // Verify that we can assign a new, non-empty configuration.
     ASSERT_NO_THROW(CfgMgr::instance().setD2ClientConfig(new_cfg));
@@ -494,18 +559,6 @@ TEST_F(CfgMgrTest, revert) {
     EXPECT_EQ(12, cfg_mgr.getCurrentCfg()->getLoggingInfo()[0].debuglevel_);
 }
 
-// This test verifies that the verbosity can be set and obtained from the
-// configuration manager.
-TEST_F(CfgMgrTest, verbosity) {
-    ASSERT_FALSE(CfgMgr::instance().isVerbose());
-
-    CfgMgr::instance().setVerbose(true);
-    ASSERT_TRUE(CfgMgr::instance().isVerbose());
-
-    CfgMgr::instance().setVerbose(false);
-    EXPECT_FALSE(CfgMgr::instance().isVerbose());
-}
-
 // This test verifies that the address family can be set and obtained
 // from the configuration manager.
 TEST_F(CfgMgrTest, family) {
@@ -558,6 +611,58 @@ TEST_F(CfgMgrTest, commitStats4) {
     EXPECT_EQ(128, total_addrs->getInteger().first);
 }
 
+// This test verifies that once the configuration is merged into the current
+// configuration, statistics are updated appropriately.
+TEST_F(CfgMgrTest, mergeIntoCurrentStats4) {
+    CfgMgr& cfg_mgr = CfgMgr::instance();
+    StatsMgr& stats_mgr = StatsMgr::instance();
+    startBackend(AF_INET);
+
+    // Let's prepare the "old" configuration: a subnet with id 123
+    // and pretend there were addresses assigned, so statistics are non-zero.
+    Subnet4Ptr subnet1(new Subnet4(IOAddress("192.1.2.0"), 24, 1, 2, 3, 123));
+    CfgSubnets4Ptr subnets = cfg_mgr.getStagingCfg()->getCfgSubnets4();
+    subnets->add(subnet1);
+    cfg_mgr.commit();
+    stats_mgr.addValue("subnet[123].total-addresses", static_cast<int64_t>(256));
+    stats_mgr.setValue("subnet[123].assigned-addresses", static_cast<int64_t>(150));
+
+    // There should be no stats for subnet 42 at this point.
+    EXPECT_FALSE(stats_mgr.getObservation("subnet[42].total-addresses"));
+    EXPECT_FALSE(stats_mgr.getObservation("subnet[42].assigned-addresses"));
+
+    // Now, let's create new configuration with updates.
+
+    // There's a subnet 192.1.3.0/24 with ID=42
+    Subnet4Ptr subnet2(new Subnet4(IOAddress("192.1.3.0"), 24, 1, 2, 3, 42));
+
+    // Let's make a pool with 128 addresses available.
+    PoolPtr pool(new Pool4(IOAddress("192.1.3.0"), 25)); // 128 addrs
+    subnet2->addPool(pool);
+
+    // Create external configuration to be merged into current one.
+    auto external_cfg = CfgMgr::instance().createExternalCfg();
+    subnets = external_cfg->getCfgSubnets4();
+    subnets->add(subnet2);
+
+    // Let's merge it.
+    cfg_mgr.mergeIntoCurrentCfg(external_cfg->getSequence());
+
+    // The stats should have been updated and so we should be able to get
+    // obeservations for subnet 42.
+    EXPECT_TRUE(stats_mgr.getObservation("subnet[42].total-addresses"));
+    EXPECT_TRUE(stats_mgr.getObservation("subnet[42].assigned-addresses"));
+
+    // And also for 123
+    EXPECT_TRUE(stats_mgr.getObservation("subnet[123].total-addresses"));
+    EXPECT_TRUE(stats_mgr.getObservation("subnet[123].assigned-addresses"));
+
+    ObservationPtr total_addrs;
+    EXPECT_NO_THROW(total_addrs = stats_mgr.getObservation("subnet[42].total-addresses"));
+    ASSERT_TRUE(total_addrs);
+    EXPECT_EQ(128, total_addrs->getInteger().first);
+}
+
 // This test verifies that once the configuration is cleared, the statistics
 // are removed.
 TEST_F(CfgMgrTest, clearStats4) {
@@ -637,6 +742,71 @@ TEST_F(CfgMgrTest, commitStats6) {
     EXPECT_EQ(65536, total_addrs->getInteger().first);
 }
 
+// This test verifies that once the configuration is merged into the current
+// configuration, statistics are updated appropriately.
+/// @todo Enable this test once merging v6 configuration is enabled.
+TEST_F(CfgMgrTest, DISABLED_mergeIntoCurrentStats6) {
+    CfgMgr& cfg_mgr = CfgMgr::instance();
+    StatsMgr& stats_mgr = StatsMgr::instance();
+    startBackend(AF_INET6);
+
+    // Let's prepare the "old" configuration: a subnet with id 123
+    // and pretend there were addresses assigned, so statistics are non-zero.
+    Subnet6Ptr subnet1(new Subnet6(IOAddress("2001:db8:1::"), 48, 1, 2, 3, 4, 123));
+    CfgSubnets6Ptr subnets = cfg_mgr.getStagingCfg()->getCfgSubnets6();
+    subnets->add(subnet1);
+    cfg_mgr.commit();
+    stats_mgr.addValue("subnet[123].total-nas", static_cast<int64_t>(256));
+    stats_mgr.setValue("subnet[123].assigned-nas", static_cast<int64_t>(150));
+
+    stats_mgr.addValue("subnet[123].total-pds", static_cast<int64_t>(256));
+    stats_mgr.setValue("subnet[123].assigned-pds", static_cast<int64_t>(150));
+
+    // There should be no stats for subnet 42 at this point.
+    EXPECT_FALSE(stats_mgr.getObservation("subnet[42].total-nas"));
+    EXPECT_FALSE(stats_mgr.getObservation("subnet[42].assigned-nas"));
+    EXPECT_FALSE(stats_mgr.getObservation("subnet[42].total-pds"));
+    EXPECT_FALSE(stats_mgr.getObservation("subnet[42].assigned-pds"));
+
+    // Now, let's create new configuration with updates.
+
+    // There's a subnet 2001:db8:2::/48 with ID=42
+    Subnet6Ptr subnet2(new Subnet6(IOAddress("2001:db8:2::"), 48, 1, 2, 3, 4, 42));
+
+    // Let's make pools with 128 addresses and 65536 prefixes available.
+    PoolPtr pool1(new Pool6(Lease::TYPE_NA, IOAddress("2001:db8:2::"), 121)); // 128 addrs
+    PoolPtr pool2(new Pool6(Lease::TYPE_PD, IOAddress("2001:db8:3::"), 96, 112)); // 65536 prefixes
+    subnet2->addPool(pool1);
+    subnet2->addPool(pool2);
+
+    // Create external configuration to be merged into current one.
+    auto external_cfg = CfgMgr::instance().createExternalCfg();
+    subnets = external_cfg->getCfgSubnets6();
+    subnets->add(subnet2);
+
+    // Let's merge it.
+    cfg_mgr.mergeIntoCurrentCfg(external_cfg->getSequence());
+
+    EXPECT_TRUE(stats_mgr.getObservation("subnet[42].total-nas"));
+    EXPECT_TRUE(stats_mgr.getObservation("subnet[42].assigned-nas"));
+    EXPECT_TRUE(stats_mgr.getObservation("subnet[42].total-pds"));
+    EXPECT_TRUE(stats_mgr.getObservation("subnet[42].assigned-pds"));
+
+    EXPECT_TRUE(stats_mgr.getObservation("subnet[123].total-nas"));
+    EXPECT_TRUE(stats_mgr.getObservation("subnet[123].assigned-nas"));
+    EXPECT_TRUE(stats_mgr.getObservation("subnet[123].total-pds"));
+    EXPECT_TRUE(stats_mgr.getObservation("subnet[123].assigned-pds"));
+
+    ObservationPtr total_addrs;
+    EXPECT_NO_THROW(total_addrs = stats_mgr.getObservation("subnet[42].total-nas"));
+    ASSERT_TRUE(total_addrs);
+    EXPECT_EQ(128, total_addrs->getInteger().first);
+
+    EXPECT_NO_THROW(total_addrs = stats_mgr.getObservation("subnet[42].total-pds"));
+    ASSERT_TRUE(total_addrs);
+    EXPECT_EQ(65536, total_addrs->getInteger().first);
+}
+
 // This test verifies that once the configuration is cleared, the v6 statistics
 // are removed.
 TEST_F(CfgMgrTest, clearStats6) {
@@ -673,6 +843,145 @@ TEST_F(CfgMgrTest, clearStats6) {
     EXPECT_FALSE(stats_mgr.getObservation("subnet[123].assigned-pds"));
 }
 
+// This test verifies that the external configuration can be merged into
+// the staging configuration via CfgMgr.
+TEST_F(CfgMgrTest, mergeIntoStagingCfg) {
+    CfgMgr& cfg_mgr = CfgMgr::instance();
+
+    // Create first external configuration.
+    SrvConfigPtr ext_cfg1;
+    ASSERT_NO_THROW(ext_cfg1 = cfg_mgr.createExternalCfg());
+    ASSERT_TRUE(ext_cfg1);
+    // It should pick the first available sequence number.
+    EXPECT_EQ(0, ext_cfg1->getSequence());
+
+    // Create second external configuration.
+    SrvConfigPtr ext_cfg2;
+    ASSERT_NO_THROW(ext_cfg2 = cfg_mgr.createExternalCfg());
+    ASSERT_TRUE(ext_cfg2);
+    // It should pick the next available sequence number.
+    EXPECT_EQ(1, ext_cfg2->getSequence());
+
+    // Those must be two separate instances.
+    ASSERT_FALSE(ext_cfg1 == ext_cfg2);
+
+    // Add a subnet which will be merged from first configuration.
+    Subnet4Ptr subnet1(new Subnet4(IOAddress("192.1.2.0"), 24, 1, 2, 3, 123));
+    ext_cfg1->getCfgSubnets4()->add(subnet1);
+
+    // Add a subnet which will be merged from the second configuration.
+    Subnet4Ptr subnet2(new Subnet4(IOAddress("192.1.3.0"), 24, 1, 2, 3, 124));
+    ext_cfg2->getCfgSubnets4()->add(subnet2);
+
+    // Merge first configuration.
+    ASSERT_NO_THROW(cfg_mgr.mergeIntoStagingCfg(ext_cfg1->getSequence()));
+    // Second attempt should fail because the configuration is discarded after
+    // the merge.
+    ASSERT_THROW(cfg_mgr.mergeIntoStagingCfg(ext_cfg1->getSequence()), BadValue);
+
+    // Check that the subnet from first configuration has been merged but not
+    // from the second configuration.
+    ASSERT_TRUE(cfg_mgr.getStagingCfg()->getCfgSubnets4()->getBySubnetId(123));
+    ASSERT_FALSE(cfg_mgr.getStagingCfg()->getCfgSubnets4()->getBySubnetId(124));
+
+    // Create another configuration instance to check what sequence it would
+    // pick. It should pick the first available one.
+    SrvConfigPtr ext_cfg3;
+    ASSERT_NO_THROW(ext_cfg3 = cfg_mgr.createExternalCfg());
+    ASSERT_TRUE(ext_cfg3);
+    EXPECT_EQ(2, ext_cfg3->getSequence());
+
+    // Merge the second and third (empty) configuration.
+    ASSERT_NO_THROW(cfg_mgr.mergeIntoStagingCfg(ext_cfg2->getSequence()));
+    ASSERT_NO_THROW(cfg_mgr.mergeIntoStagingCfg(ext_cfg3->getSequence()));
+
+    // Make sure that both subnets have been merged.
+    ASSERT_TRUE(cfg_mgr.getStagingCfg()->getCfgSubnets4()->getBySubnetId(123));
+    ASSERT_TRUE(cfg_mgr.getStagingCfg()->getCfgSubnets4()->getBySubnetId(124));
+
+    // The next configuration instance should reset the sequence to 0 because
+    // there are no other configurations in CfgMgr.
+    SrvConfigPtr ext_cfg4;
+    ASSERT_NO_THROW(ext_cfg4 = cfg_mgr.createExternalCfg());
+    ASSERT_TRUE(ext_cfg4);
+    EXPECT_EQ(0, ext_cfg4->getSequence());
+
+    // Try to commit the staging configuration.
+    ASSERT_NO_THROW(cfg_mgr.commit());
+
+    // Make sure that both subnets are present in the current configuration.
+    EXPECT_TRUE(cfg_mgr.getCurrentCfg()->getCfgSubnets4()->getBySubnetId(123));
+    EXPECT_TRUE(cfg_mgr.getCurrentCfg()->getCfgSubnets4()->getBySubnetId(124));
+
+    // The staging configuration should not include them.
+    EXPECT_FALSE(cfg_mgr.getStagingCfg()->getCfgSubnets4()->getBySubnetId(123));
+    EXPECT_FALSE(cfg_mgr.getStagingCfg()->getCfgSubnets4()->getBySubnetId(124));
+}
+
+// This test verifies that the external configuration can be merged into
+// the current configuration via CfgMgr.
+TEST_F(CfgMgrTest, mergeIntoCurrentCfg) {
+    CfgMgr& cfg_mgr = CfgMgr::instance();
+
+    // Create first external configuration.
+    SrvConfigPtr ext_cfg1;
+    ASSERT_NO_THROW(ext_cfg1 = cfg_mgr.createExternalCfg());
+    ASSERT_TRUE(ext_cfg1);
+    // It should pick the first available sequence number.
+    EXPECT_EQ(0, ext_cfg1->getSequence());
+
+    // Create second external configuration.
+    SrvConfigPtr ext_cfg2;
+    ASSERT_NO_THROW(ext_cfg2 = cfg_mgr.createExternalCfg());
+    ASSERT_TRUE(ext_cfg2);
+    // It should pick the next available sequence number.
+    EXPECT_EQ(1, ext_cfg2->getSequence());
+
+    // Those must be two separate instances.
+    ASSERT_FALSE(ext_cfg1 == ext_cfg2);
+
+    // Add a subnet which will be merged from first configuration.
+    Subnet4Ptr subnet1(new Subnet4(IOAddress("192.1.2.0"), 24, 1, 2, 3, 123));
+    ext_cfg1->getCfgSubnets4()->add(subnet1);
+
+    // Add a subnet which will be merged from the second configuration.
+    Subnet4Ptr subnet2(new Subnet4(IOAddress("192.1.3.0"), 24, 1, 2, 3, 124));
+    ext_cfg2->getCfgSubnets4()->add(subnet2);
+
+    // Merge first configuration.
+    ASSERT_NO_THROW(cfg_mgr.mergeIntoCurrentCfg(ext_cfg1->getSequence()));
+    // Second attempt should fail because the configuration is discarded after
+    // the merge.
+    ASSERT_THROW(cfg_mgr.mergeIntoCurrentCfg(ext_cfg1->getSequence()), BadValue);
+
+    // Check that the subnet from first configuration has been merged but not
+    // from the second configuration.
+    ASSERT_TRUE(cfg_mgr.getCurrentCfg()->getCfgSubnets4()->getBySubnetId(123));
+    ASSERT_FALSE(cfg_mgr.getCurrentCfg()->getCfgSubnets4()->getBySubnetId(124));
+
+    // Create another configuration instance to check what sequence it would
+    // pick. It should pick the first available one.
+    SrvConfigPtr ext_cfg3;
+    ASSERT_NO_THROW(ext_cfg3 = cfg_mgr.createExternalCfg());
+    ASSERT_TRUE(ext_cfg3);
+    EXPECT_EQ(2, ext_cfg3->getSequence());
+
+    // Merge the second and third (empty) configuration.
+    ASSERT_NO_THROW(cfg_mgr.mergeIntoCurrentCfg(ext_cfg2->getSequence()));
+    ASSERT_NO_THROW(cfg_mgr.mergeIntoCurrentCfg(ext_cfg3->getSequence()));
+
+    // Make sure that both subnets have been merged.
+    ASSERT_TRUE(cfg_mgr.getCurrentCfg()->getCfgSubnets4()->getBySubnetId(123));
+    ASSERT_TRUE(cfg_mgr.getCurrentCfg()->getCfgSubnets4()->getBySubnetId(124));
+
+    // The next configuration instance should reset the sequence to 0 because
+    // there are no other configurations in CfgMgr.
+    SrvConfigPtr ext_cfg4;
+    ASSERT_NO_THROW(ext_cfg4 = cfg_mgr.createExternalCfg());
+    ASSERT_TRUE(ext_cfg4);
+    EXPECT_EQ(0, ext_cfg4->getSequence());
+}
+
 /// @todo Add unit-tests for testing:
 /// - addActiveIface() with invalid interface name
 /// - addActiveIface() with the same interface twice
