@@ -1,4 +1,4 @@
-// Copyright (C) 2014-2017 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2014-2019 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -10,16 +10,17 @@
 #include <dhcpsrv/lease_mgr.h>
 #include <gtest/gtest.h>
 #include <vector>
+#include <set>
 
 namespace isc {
 namespace dhcp {
 namespace test {
 
-
 /// @brief typedefs to simplify lease statistic testing
 typedef std::map<std::string, int64_t> StatValMap;
 typedef std::pair<std::string, int64_t> StatValPair;
 typedef std::vector<StatValMap> StatValMapList;
+typedef std::set<LeaseStatsRow> RowSet;
 
 /// @brief Test Fixture class with utility functions for LeaseMgr backends
 ///
@@ -27,12 +28,8 @@ typedef std::vector<StatValMap> StatValMapList;
 /// All concrete LeaseMgr test classes should be derived from it.
 class GenericLeaseMgrTest : public ::testing::Test {
 public:
-
     /// @brief Universe (V4 or V6).
-    enum Universe {
-        V4,
-        V6
-    };
+    enum Universe { V4, V6 };
 
     /// @brief Default constructor.
     GenericLeaseMgrTest();
@@ -70,7 +67,8 @@ public:
     ///
     /// @param address Address to use for the initialization
     ///
-    /// @return Lease6Ptr.  This will not point to anything if the initialization
+    /// @return Lease6Ptr.  This will not point to anything if the
+    /// initialization
     ///         failed (e.g. unknown address).
     Lease6Ptr initializeLease6(std::string address);
 
@@ -165,10 +163,10 @@ public:
     /// @brief Test lease retrieval using client id, HW address and subnet id.
     void testGetLease4ClientIdHWAddrSubnetId();
 
-    // @brief Get lease4 by hardware address (2)
-    //
-    // Check that the system can cope with getting a hardware address of
-    // any size.
+    /// @brief Get lease4 by hardware address (2)
+    ///
+    /// Check that the system can cope with getting a hardware address of
+    /// any size.
     void testGetLease4HWAddrSize();
 
     /// @brief Check GetLease4 methods - access by Hardware Address & Subnet ID
@@ -200,6 +198,24 @@ public:
     /// a combination of client and subnet IDs.
     void testGetLease4ClientIdSubnetId();
 
+    /// @brief Test method which returns all IPv4 leases for Subnet ID.
+    void testGetLeases4SubnetId();
+
+    /// @brief Test method which returns all IPv4 leases.
+    void testGetLeases4();
+
+    /// @brief Test method which returns range of IPv4 leases with paging.
+    void testGetLeases4Paged();
+
+    /// @brief Test method which returns all IPv6 leases for Subnet ID.
+    void testGetLeases6SubnetId();
+
+    /// @brief Test method which returns all IPv6 leases.
+    void testGetLeases6();
+
+    /// @brief Test method which returns range of IPv6 leases with paging.
+    void testGetLeases6Paged();
+
     /// @brief Basic Lease4 Checks
     ///
     /// Checks that the addLease, getLease4(by address), getLease4(hwaddr,subnet_id),
@@ -263,6 +279,12 @@ public:
     /// Adds leases to the database and checks that they can be accessed via
     /// a combination of DIUID and IAID.
     void testGetLease6DuidIaidSubnetId();
+    
+    /// @brief verifies getLeases6 method by DUID
+    ///
+    /// Adds 3 leases to backend and retrieves, verifes empty
+    /// retrival of non existent DUID.
+    void testGetLeases6Duid();
 
     /// @brief Checks that getLease6() works with different DUID sizes
     void testGetLease6DuidIaidSubnetIdSize();
@@ -387,14 +409,51 @@ public:
     /// attempts to delete them, one subnet at a time.
     void testWipeLeases6();
 
+    /// @brief Checks operation of v4 LeaseStatsQuery variants
+    ///
+    /// It creates three subnets with leasese in various states in
+    /// each.  It runs and verifies the returned query contents for
+    /// each of the v4 startLeaseQuery variants:
+    ///
+    /// - startSubnetLeaseQuery()
+    /// - startSubneRangetLeaseQuery()
+    /// - startLeaseQuery()
+    ///
+    void testLeaseStatsQuery4();
+
+    /// @brief Checks operation of v6 LeaseStatsQuery variants
+    ///
+    /// It creates three subnets with leasese in various states in
+    /// each.  It runs and verifies the returned query contents for
+    /// each of the v6 startLeaseQuery variants:
+    ///
+    /// - startSubnetLeaseQuery()
+    /// - startSubneRangetLeaseQuery()
+    /// - startLeaseQuery()
+    ///
+    void testLeaseStatsQuery6();
+
+    /// @brief Compares LeaseQueryStats content to expected set of rows
+    ///
+    /// @param qry - a started LeaseStatsQuery
+    /// @param row_set - set of rows expected to be found in the query rows
+    void checkQueryAgainstRowSet(const LeaseStatsQueryPtr& qry, const RowSet& row_set);
+
+    /// @brief Checks if specified range of leases was returned.
+    ///
+    /// @param returned collection of leases returned.
+    /// @param expected_addresses ordered collection of expected addresses.
+    void checkLeaseRange(const Lease4Collection& returned,
+                         const std::vector<std::string>& expected_addresses);
+
     /// @brief String forms of IPv4 addresses
-    std::vector<std::string>  straddress4_;
+    std::vector<std::string> straddress4_;
 
     /// @brief IOAddress forms of IPv4 addresses
     std::vector<isc::asiolink::IOAddress> ioaddress4_;
 
     /// @brief String forms of IPv6 addresses
-    std::vector<std::string>  straddress6_;
+    std::vector<std::string> straddress6_;
 
     /// @brief Types of IPv6 Leases
     std::vector<Lease::Type> leasetype6_;
@@ -406,8 +465,73 @@ public:
     LeaseMgr* lmptr_;
 };
 
-}; // namespace test
-}; // namespace dhcp
-}; // namespace isc
+class LeaseMgrDbLostCallbackTest : public ::testing::Test {
+public:
+    LeaseMgrDbLostCallbackTest() {
+        db::DatabaseConnection::db_lost_callback = 0;
+    }
+
+    virtual ~LeaseMgrDbLostCallbackTest() {
+        db::DatabaseConnection::db_lost_callback = 0;
+    }
+
+    /// @brief Prepares the class for a test.
+    ///
+    /// Invoked by gtest prior test entry, we create the
+    /// appropriate schema and wipe out any residual lease manager
+    virtual void SetUp();
+
+    /// @brief Pre-text exit clean up
+    ///
+    /// Invoked by gtest upon test exit, we destroy the schema
+    /// we created and toss our lease manager.
+    virtual void TearDown();
+
+    /// @brief Abstract method for destroying the back end specific shcema
+    virtual void destroySchema() = 0;
+
+    /// @brief Abstract method for creating the back end specific shcema
+    virtual void createSchema() = 0;
+
+    /// @brief Abstract method which returns the back end specific connection
+    /// string
+    virtual std::string validConnectString() = 0;
+
+    /// @brief Abstract method which returns invalid back end specific connection
+    /// string
+    virtual std::string invalidConnectString() = 0;
+
+    /// @brief Verifies open failures do NOT invoke db lost callback
+    ///
+    /// The db lost callback should only be invoked after successfully
+    /// opening the DB and then subsequently losing it. Failing to
+    /// open should be handled directly by the application layer.
+    void testNoCallbackOnOpenFailure();
+
+    /// @brief Verifies the host manager's behavior if DB connection is lost
+    ///
+    /// This function creates a lease manager with an back end that
+    /// supports connectivity lost callback (currently only MySQL and
+    /// PostgreSQL currently).  It verifies connectivity by issuing a known
+    /// valid query.  Next it simulates connectivity lost by identifying and
+    /// closing the socket connection to the host backend.  It then reissues
+    /// the query and verifies that:
+    /// -# The Query throws  DbOperationError (rather than exiting)
+    /// -# The registered DbLostCallback was invoked
+    void testDbLostCallback();
+
+    /// @brief Callback function registered with the host manager
+    bool db_lost_callback(db::ReconnectCtlPtr /* not_used */) {
+        return (callback_called_ = true);
+    }
+
+    /// @brief Flag used to detect calls to db_lost_callback function
+    bool callback_called_;
+
+};
+
+}  // namespace test
+}  // namespace dhcp
+}  // namespace isc
 
 #endif
