@@ -1,4 +1,4 @@
-// Copyright (C) 2015-2017 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2015-2018 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -390,6 +390,54 @@ public:
         checkTokenRelay6Field(eval.expression.at(0), exp_level, exp_type);
     }
 
+    /// @brief checks if the given token is a TokenMember with the
+    /// correct client class name.
+    /// @param token token to be checked
+    /// @param expected_client_class expected client class name
+    void checkTokenMember(const TokenPtr& token,
+                          const std::string& expected_client_class) {
+        ASSERT_TRUE(token);
+        boost::shared_ptr<TokenMember> member =
+            boost::dynamic_pointer_cast<TokenMember>(token);
+        ASSERT_TRUE(member);
+
+        EXPECT_EQ(expected_client_class, member->getClientClass());
+    }
+
+    /// @rief This tests attempts to parse the expression then checks          
+    /// if the number of tokens is correct and the TokenMember is as
+    /// expected.
+    ///
+    /// @param expr expression to be parsed
+    /// @param check_defined closure checking if the client class is defined
+    /// @param exp_client_class expected client class name to be parsed
+    /// @param exp_tokens expected number of tokens
+    void testMember(const std::string& expr,
+                    EvalContext::CheckDefined check_defined,
+                    const std::string& exp_client_class,
+                    int exp_tokens) {
+        EvalContext eval(Option::V6, check_defined);
+
+        // parse the expression
+        try {
+            parsed_ = eval.parseString(expr);
+        }
+        catch (const EvalParseError& ex) {
+            FAIL() <<"Exception thrown: " << ex.what();
+            return;
+        }
+
+        // Parsing should succeed and return a token.
+        EXPECT_TRUE(parsed_);
+
+        // There should be the expected number of tokens.
+        ASSERT_EQ(exp_tokens, eval.expression.size());
+
+        // checked that the first token is TokenRelay6Field and that
+        // is has the correct attributes
+        checkTokenMember(eval.expression.at(0), exp_client_class);
+    }
+
     /// @brief checks if the given token is a substring operator
     void checkTokenSubstring(const TokenPtr& token) {
         ASSERT_TRUE(token);
@@ -406,6 +454,22 @@ public:
         EXPECT_TRUE(conc);
     }
 
+    /// @brief checks if the given token is an ifelse operator
+    void checkTokenIfElse(const TokenPtr& token) {
+        ASSERT_TRUE(token);
+        boost::shared_ptr<TokenIfElse> alt =
+            boost::dynamic_pointer_cast<TokenIfElse>(token);
+        EXPECT_TRUE(alt);
+    }
+
+    /// @brief checks if the given token is a hexstring operator
+    void checkTokenToHexString(const TokenPtr& token) {
+        ASSERT_TRUE(token);
+        boost::shared_ptr<TokenToHexString> tohex =
+            boost::dynamic_pointer_cast<TokenToHexString>(token);
+        EXPECT_TRUE(tohex);
+    }
+
     /// @brief checks if the given expression raises the expected message
     /// when it is parsed.
     void checkError(const string& expr, const string& msg) {
@@ -1000,6 +1064,12 @@ TEST_F(EvalContextTest, pkt4FieldChaddr) {
     testPkt4Field("pkt4.mac == 0x000102030405", TokenPkt4::CHADDR, 3);
 }
 
+// Tests whether chaddr field in DHCPv4 can be accessed and converted.
+TEST_F(EvalContextTest, pkt4FieldChaddrHexa) {
+    testPkt4Field("hexstring(pkt4.mac, ':') == '00:01:02:03:04:05'",
+                  TokenPkt4::CHADDR, 5);
+}
+
 // Tests whether hlen field in DHCPv4 can be accessed.
 TEST_F(EvalContextTest, pkt4FieldHlen) {
     testPkt4Field("pkt4.hlen == 0x6", TokenPkt4::HLEN, 3);
@@ -1063,6 +1133,31 @@ TEST_F(EvalContextTest, relay6FieldError) {
                "<string>:1.8: Nest level invalid for DHCPv4 packets");
 }
 
+// Tests parsing of member with defined class
+TEST_F(EvalContextTest, member) {
+    auto check_defined = [](const ClientClass& cc) { return (cc == "foo"); };
+    testMember("member('foo')", check_defined, "foo", 1);
+}
+
+// Test parsing of member with not defined class
+TEST_F(EvalContextTest, memberError) {
+    auto check_defined = [](const ClientClass& cc) { return (cc == "foo"); };
+    EvalContext eval(Option::V6, check_defined);
+    parsed_ = false;
+    try {
+        parsed_ = eval.parseString("member('bar')");
+        FAIL() << "Expected EvalParseError but nothing was raised";
+    }
+    catch (const EvalParseError& ex) {
+        EXPECT_EQ("<string>:1.8-12: Not defined client class 'bar'",
+                  std::string(ex.what()));
+        EXPECT_FALSE(parsed_);
+    }
+    catch (...) {
+        FAIL() << "Expected EvalParseError but something else was raised";
+    }
+}
+
 // Test parsing of logical operators
 TEST_F(EvalContextTest, logicalOps) {
     // option.exists
@@ -1209,6 +1304,45 @@ TEST_F(EvalContextTest, concat) {
     checkTokenConcat(tmp3);
 }
 
+// Test the parsing of an ifelse expression
+TEST_F(EvalContextTest, ifElse) {
+    EvalContext eval(Option::V4);
+
+    EXPECT_NO_THROW(parsed_ =
+        eval.parseString("ifelse('foo' == 'bar', 'us', 'them') == 'you'"));
+
+    ASSERT_EQ(8, eval.expression.size());
+
+    TokenPtr tmp1 = eval.expression.at(2);
+    TokenPtr tmp2 = eval.expression.at(3);
+    TokenPtr tmp3 = eval.expression.at(4);
+    TokenPtr tmp4 = eval.expression.at(5);
+
+    checkTokenEq(tmp1);
+    checkTokenString(tmp2, "us");
+    checkTokenString(tmp3, "them");
+    checkTokenIfElse(tmp4);
+}
+
+// Test the parsing of a hexstring expression
+TEST_F(EvalContextTest, toHexString) {
+    EvalContext eval(Option::V4);
+
+    EXPECT_NO_THROW(parsed_ =
+        eval.parseString("hexstring(0x666f,'-') == '66-6f'"));
+    EXPECT_TRUE(parsed_);
+
+    ASSERT_EQ(5, eval.expression.size());
+
+    TokenPtr tmp1 = eval.expression.at(0);
+    TokenPtr tmp2 = eval.expression.at(1);
+    TokenPtr tmp3 = eval.expression.at(2);
+
+    checkTokenHexString(tmp1, "fo");
+    checkTokenString(tmp2, "-");
+    checkTokenToHexString(tmp3);
+}
+
 //
 // Test some scanner error cases
 TEST_F(EvalContextTest, scanErrors) {
@@ -1221,11 +1355,11 @@ TEST_F(EvalContextTest, scanErrors) {
 
     // Typo should be handled as well.
     checkError("subtring", "<string>:1.1: Invalid character: s");
-
     checkError("foo", "<string>:1.1: Invalid character: f");
     checkError(" bar", "<string>:1.2: Invalid character: b");
     checkError("relay[12].hex == 'foo'", "<string>:1.1: Invalid character: r");
     checkError("pkt4.ziaddr", "<string>:1.6: Invalid character: z");
+    checkError("members('foo'", "<string>:1.7: Invalid character: s");
 }
 
 // Tests some scanner/parser error cases
@@ -1358,6 +1492,10 @@ TEST_F(EvalContextTest, parseErrors) {
                "<string>:1.16: syntax error, unexpected ), expecting \",\"");
     checkError("concat('foo','bar','') == 'foobar'",
                "<string>:1.19: syntax error, unexpected \",\", expecting )");
+    checkError("ifelse('foo'=='bar','foo')",
+               "<string>:1.26: syntax error, unexpected ), expecting \",\"");
+    checkError("ifelse('foo'=='bar','foo','bar','')",
+               "<string>:1.32: syntax error, unexpected \",\", expecting )");
 }
 
 // Tests some type error cases
@@ -1388,6 +1526,17 @@ TEST_F(EvalContextTest, typeErrors) {
                "<string>:1.8-10: syntax error, unexpected and, expecting ==");
     checkError("'true' or 'false'",
                "<string>:1.8-9: syntax error, unexpected or, expecting ==");
+    // Ifelse requires a boolean condition and string branches.
+    checkError("ifelse('foobar','foo','bar')",
+               "<string>:1.16: syntax error, unexpected \",\", expecting ==");
+    checkError("ifelse('foo'=='bar','foo'=='foo','bar')",
+               "<string>:1.26-27: syntax error, unexpected ==, "
+               "expecting \",\"");
+    checkError("ifelse('foo'=='bar','foo','bar'=='bar')",
+               "<string>:1.32-33: syntax error, unexpected ==, expecting )");
+
+    // Member uses quotes around the client class name.
+    checkError("member(foo)", "<string>:1.8: Invalid character: f");
 }
 
 
