@@ -1,4 +1,4 @@
-// Copyright (C) 2013-2017 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2013-2019 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -7,25 +7,15 @@
 #include <config.h>
 #include <cc/command_interpreter.h>
 #include <cfgrpt/config_report.h>
-#include <cryptolink/cryptolink.h>
-#include <dhcpsrv/cfgmgr.h>
 #include <exceptions/exceptions.h>
 #include <log/logger.h>
 #include <log/logger_support.h>
+#include <process/daemon.h>
 #include <process/d_log.h>
 #include <process/d_controller.h>
-
-#ifdef HAVE_MYSQL
-#include <dhcpsrv/mysql_lease_mgr.h>
-#endif
-#ifdef HAVE_PGSQL
-#include <dhcpsrv/pgsql_lease_mgr.h>
-#endif
-#ifdef HAVE_CQL
-#include <dhcpsrv/cql_lease_mgr.h>
-#endif
-#include <dhcpsrv/memfile_lease_mgr.h>
-
+#include <process/config_base.h>
+#include <kea_version.h>
+#include <boost/bind.hpp>
 #include <sstream>
 #include <unistd.h>
 
@@ -40,7 +30,7 @@ DControllerBasePtr DControllerBase::controller_;
 // Note that the constructor instantiates the controller's primary IOService.
 DControllerBase::DControllerBase(const char* app_name, const char* bin_name)
     : app_name_(app_name), bin_name_(bin_name),
-      verbose_(false), check_only_(false), spec_file_name_(""),
+      verbose_(false), check_only_(false),
       io_service_(new isc::asiolink::IOService()),
       io_signal_queue_() {
 }
@@ -85,12 +75,12 @@ DControllerBase::launch(int argc, char* argv[], const bool test_mode) {
     // It is important that we set a default logger name because this name
     // will be used when the user doesn't provide the logging configuration
     // in the Kea configuration file.
-    isc::dhcp::CfgMgr::instance().setDefaultLoggerName(bin_name_);
+    Daemon::setDefaultLoggerName(bin_name_);
 
     // Logger's default configuration depends on whether we are in the
     // verbose mode or not. CfgMgr manages the logger configuration so
     // the verbose mode is set for CfgMgr.
-    isc::dhcp::CfgMgr::instance().setVerbose(verbose_);
+    Daemon::setVerbose(verbose_);
 
     // Do not initialize logger here if we are running unit tests. It would
     // replace an instance of unit test specific logger.
@@ -99,9 +89,17 @@ DControllerBase::launch(int argc, char* argv[], const bool test_mode) {
         Daemon::loggerInit(bin_name_.c_str(), verbose_);
     }
 
+    try {
+        checkConfigFile();
+    } catch (const std::exception& ex) {
+        LOG_FATAL(dctl_logger, DCTL_CONFIG_FILE_LOAD_FAIL)
+            .arg(app_name_).arg(ex.what());
+        isc_throw (LaunchError, "Launch Failed: " << ex.what());
+    }
+
     try {
         createPIDFile();
-    } catch (const dhcp::DaemonPIDExists& ex) {
+    } catch (const DaemonPIDExists& ex) {
         LOG_FATAL(dctl_logger, DCTL_ALREADY_RUNNING)
                   .arg(bin_name_).arg(ex.what());
         isc_throw (LaunchError, "Launch Failed: " << ex.what());
@@ -162,8 +160,8 @@ DControllerBase::checkConfigOnly() {
         // messages are to be printed.
         // This is just a test, so we don't care about lockfile.
         setenv("KEA_LOCKFILE_DIR", "none", 0);
-        isc::dhcp::CfgMgr::instance().setDefaultLoggerName(bin_name_);
-        isc::dhcp::CfgMgr::instance().setVerbose(verbose_);
+        Daemon::setDefaultLoggerName(bin_name_);
+        Daemon::setVerbose(verbose_);
         Daemon::loggerInit(bin_name_.c_str(), verbose_);
 
         // Check the syntax first.
@@ -188,6 +186,33 @@ DControllerBase::checkConfigOnly() {
             isc_throw(InvalidUsage, "Config file " << config_file <<
                       " does not include '" << getAppName() << "' entry");
         }
+        if (module_config->getType() != Element::map) {
+            isc_throw(InvalidUsage, "Config file " << config_file <<
+                      " include not map '" << getAppName() << "' entry");
+        }
+
+        // Check obsolete or unknown (aka unsupported) objects.
+        for (auto obj : whole_config->mapValue()) {
+            const std::string& app_name = getAppName();
+            const std::string& obj_name = obj.first;
+            if (obj_name == app_name) {
+                continue;
+            }
+            if (obj_name == "Logging") {
+                LOG_WARN(dctl_logger, DCTL_CONFIG_DEPRECATED)
+                    .arg("'Logging' defined in top level. This is deprecated."
+                         " Please define it in the '" + app_name + "' scope.");
+                continue;
+            }
+            LOG_WARN(dctl_logger, DCTL_CONFIG_DEPRECATED)
+                .arg("'" + obj_name + "', defining anything in global level besides '"
+                     + app_name + "' is no longer supported.");
+        }
+
+        // Relocate Logging: if there is a global Logging object takes its
+        // loggers entry, move the entry to AppName object and remove
+        // now empty Logging.
+        Daemon::relocateLogging(whole_config, getAppName());
 
         // Get an application process object.
         initProcess();
@@ -212,6 +237,11 @@ DControllerBase::checkConfigOnly() {
 void
 DControllerBase::parseArgs(int argc, char* argv[])
 {
+
+    if (argc == 1) {
+        isc_throw(InvalidUsage, "");
+    }
+
     // Iterate over the given command line options. If its a stock option
     // ("c" or "d") handle it here.  If its a valid custom option, then
     // invoke customOption.
@@ -319,7 +349,9 @@ ConstElementPtr
 DControllerBase::configFromFile() {
     // Rollback any previous staging configuration. For D2, only a
     // logger configuration is used here.
-    isc::dhcp::CfgMgr::instance().rollback();
+    // We're not using cfgmgr to store logging configuration anymore.
+    // isc::dhcp::CfgMgr::instance().rollback();
+
     // Will hold configuration.
     ConstElementPtr module_config;
     // Will receive configuration result.
@@ -340,17 +372,6 @@ DControllerBase::configFromFile() {
             whole_config = Element::fromJSONFile(config_file, true);
         }
 
-        // Let's configure logging before applying the configuration,
-        // so we can log things during configuration process.
-
-        // Temporary storage for logging configuration
-        isc::dhcp::SrvConfigPtr storage =
-            isc::dhcp::CfgMgr::instance().getStagingCfg();
-
-        // Get 'Logging' element from the config and use it to set up
-        // logging. If there's no such element, we'll just pass NULL.
-        Daemon::configureLogger(whole_config->get("Logging"), storage);
-
         // Extract derivation-specific portion of the configuration.
         module_config = whole_config->get(getAppName());
         if (!module_config) {
@@ -358,20 +379,56 @@ DControllerBase::configFromFile() {
                                 " does not include '" <<
                                  getAppName() << "' entry.");
         }
-
-        answer = updateConfig(module_config);
-        int rcode = 0;
-        parseAnswer(rcode, answer);
-        if (!rcode) {
-            // Configuration successful, so apply the logging configuration
-            // to log4cplus.
-            isc::dhcp::CfgMgr::instance().getStagingCfg()->applyLoggingCfg();
-            isc::dhcp::CfgMgr::instance().commit();
+        if (module_config->getType() != Element::map) {
+            isc_throw(InvalidUsage, "Config file " << config_file <<
+                      " include not map '" << getAppName() << "' entry");
         }
 
+        // Check obsolete or unknown (aka unsupported) objects.
+        for (auto obj : whole_config->mapValue()) {
+            const std::string& app_name = getAppName();
+            const std::string& obj_name = obj.first;
+            if (obj_name == app_name) {
+                continue;
+            }
+            if (obj_name == "Logging") {
+                LOG_WARN(dctl_logger, DCTL_CONFIG_DEPRECATED)
+                    .arg("Logging defined in top level. This is deprecated."
+                         " Please define it in the '" + app_name + "' scope.");
+                continue;
+            }
+            LOG_WARN(dctl_logger, DCTL_CONFIG_DEPRECATED)
+                .arg("'" + obj_name + "', defining anything in global level besides '"
+                     + app_name + "' is no longer supported.");
+        }
+
+        // Relocate Logging: if there is a global Logging object takes its
+        // loggers entry, move the entry to AppName object and remove
+        // now empty Logging.
+        Daemon::relocateLogging(whole_config, getAppName());
+
+        // Let's configure logging before applying the configuration,
+        // so we can log things during configuration process.
+
+        // Temporary storage for logging configuration
+        ConfigPtr storage(new ConfigBase());
+
+        // Get 'Logging' element from the config and use it to set up
+        // logging. If there's no such element, we'll just pass NULL.
+        Daemon::configureLogger(module_config, storage);
+
+        // Let's apply the new logging. We do it early, so we'll be able
+        // to print out what exactly is wrong with the new config in
+        // case of problems.
+        storage->applyLoggingCfg();
+
+        answer = updateConfig(module_config);
+        // In all cases the right logging configuration is in the context.
+        process_->getCfgMgr()->getContext()->applyLoggingCfg();
     } catch (const std::exception& ex) {
         // Rollback logging configuration.
-        isc::dhcp::CfgMgr::instance().rollback();
+        process_->getCfgMgr()->getContext()->applyLoggingCfg();
+
         // build an error result
         ConstElementPtr error = createAnswer(COMMAND_ERROR,
                  std::string("Configuration parsing failed: ") + ex.what());
@@ -381,7 +438,6 @@ DControllerBase::configFromFile() {
     return (answer);
 }
 
-
 void
 DControllerBase::runProcess() {
     LOG_DEBUG(dctl_logger, isc::log::DBGLVL_START_SHUT, DCTL_RUN_PROCESS)
@@ -447,30 +503,37 @@ DControllerBase::configWriteHandler(const std::string&,
         }
     }
 
-    // Now do the sanity checks on the filename
-    if (filename.find("..") != std::string::npos) {
-        // Trying to escape the directory.. nope.
-        return (createAnswer(COMMAND_ERROR,
-                             "Using '..' in filename is not allowed."));
-    }
-
-    if (filename.find("\\") != std::string::npos) {
-        // Trying to inject escapes (possibly to inject quotes and something
-        // nasty afterward)
-        return (createAnswer(COMMAND_ERROR,
-                             "Using \\ in filename is not allowed."));
-    }
-
-    if (filename[0] == '/') {
-        // Absolute paths are not allowed.
-        return (createAnswer(COMMAND_ERROR,
-                             "Absolute path in filename is not allowed."));
-    }
 
     // Ok, it's time to write the file.
     size_t size = 0;
+    ElementPtr cfg = process_->getCfgMgr()->getContext()->toElement();
+
+    // Logging storage is messed up in CA. During its configuration (see
+    // DControllerBase::configFromFile() it calls Daemon::configureLogger()
+    // that stores the logging info in isc::dhcp::CfgMgr::getStagingCfg().
+    // This is later moved to getCurrentCfg() when the configuration is
+    // commited. All control-agent specific configuration is stored in
+    // a structure accessible by process_->getCfgMgr()->getContext(). Note
+    // logging information is not stored there.
+    //
+    // As a result, we need to extract the CA configuration from one
+    // place and logging from another.
+    if (!cfg->contains("Logging")) {
+        ConfigPtr base_cfg = process_->getCfgMgr()->getContext();
+
+        ConstElementPtr loginfo = base_cfg->toElement();
+        if (loginfo) {
+            // If there was a config stored in dhcp::CfgMgr, try to get Logging info from it.
+            loginfo = loginfo->get("Logging");
+        }
+        if (loginfo) {
+            // If there is some logging information, add it to our config.
+            cfg->set("Logging", loginfo);
+        }
+    }
+
     try {
-        size = writeConfigFile(filename);
+        size = writeConfigFile(filename, cfg);
     } catch (const isc::Exception& ex) {
         return (createAnswer(COMMAND_ERROR,
                              std::string("Error during write-config:")
@@ -522,12 +585,74 @@ DControllerBase::configTestHandler(const std::string&, ConstElementPtr args) {
     // We are starting the configuration process so we should remove any
     // staging configuration that has been created during previous
     // configuration attempts.
-    isc::dhcp::CfgMgr::instance().rollback();
+    // We're not using cfgmgr to store logging information anymore.
+    // isc::dhcp::CfgMgr::instance().rollback();
 
     // Now we check the server proper.
     return (checkConfig(module_config));
 }
 
+ConstElementPtr
+DControllerBase::configReloadHandler(const std::string&, ConstElementPtr) {
+    // Add reload in message?
+    return (configFromFile());
+}
+
+ConstElementPtr
+DControllerBase::configSetHandler(const std::string&, ConstElementPtr args) {
+    const int status_code = COMMAND_ERROR; // 1 indicates an error
+    ConstElementPtr module_config;
+    std::string app_name = getAppName();
+    std::string message;
+
+    // Command arguments are expected to be:
+    // { "Module": { ... }, "Logging": { ... } }
+    // The Logging component is technically optional. If it's not supplied
+    // logging will revert to default logging.
+    if (!args) {
+        message = "Missing mandatory 'arguments' parameter.";
+    } else {
+      module_config = args->get(app_name);
+        if (!module_config) {
+            message = "Missing mandatory '" + app_name + "' parameter.";
+        } else if (module_config->getType() != Element::map) {
+            message = "'" + app_name + "' parameter expected to be a map.";
+        }
+    }
+
+    if (!message.empty()) {
+        // Something is amiss with arguments, return a failure response.
+        ConstElementPtr result = isc::config::createAnswer(status_code,
+                                                           message);
+        return (result);
+    }
+
+    // We are starting the configuration process so we should remove any
+    // staging configuration that has been created during previous
+    // configuration attempts.
+    // We're not using cfgmgr to store logging information anymore.
+    // isc::dhcp::CfgMgr::instance().rollback();
+
+    // Temporary storage for logging configuration
+    ConfigPtr storage = process_->getCfgMgr()->getContext();
+
+    // Get 'Logging' element from the config and use it to set up
+    // logging. If there's no such element, we'll just pass NULL.
+    Daemon::configureLogger(args->get("Logging"), storage);
+
+    // Now we check the server proper.
+    ConstElementPtr answer = updateConfig(module_config);
+    int rcode = 0;
+    parseAnswer(rcode, answer);
+    if (!rcode) {
+        // Configuration successful, so apply the logging configuration
+        // to log4cplus.
+        storage->applyLoggingCfg();
+    }
+
+    return (answer);
+}
+
 ConstElementPtr
 DControllerBase::versionGetHandler(const std::string&, ConstElementPtr) {
     ConstElementPtr answer;
@@ -674,20 +799,7 @@ DControllerBase::getVersion(bool extended) {
         tmp << std::endl << EXTENDED_VERSION << std::endl;
         tmp << "linked with:" << std::endl;
         tmp << isc::log::Logger::getVersion() << std::endl;
-        tmp << isc::cryptolink::CryptoLink::getVersion() << std::endl;
-        tmp << "database:" << std::endl;
-#ifdef HAVE_MYSQL
-        tmp << isc::dhcp::MySqlLeaseMgr::getDBVersion() << std::endl;
-#endif
-#ifdef HAVE_PGSQL
-        tmp << isc::dhcp::PgSqlLeaseMgr::getDBVersion() << std::endl;
-#endif
-#ifdef HAVE_CQL
-        tmp << isc::dhcp::CqlLeaseMgr::getDBVersion() << std::endl;
-#endif
-        tmp << isc::dhcp::Memfile_LeaseMgr::getDBVersion();
-
-        // @todo: more details about database runtime
+        tmp << getVersionAddendum();
     }
 
     return (tmp.str());
