@@ -1,4 +1,4 @@
-// Copyright (C) 2011-2016 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2011-2019 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -28,6 +28,7 @@
 #include <boost/scoped_ptr.hpp>
 
 #include <gtest/gtest.h>
+#include <testutils/sandbox.h>
 
 #include <exceptions/exceptions.h>
 
@@ -147,17 +148,17 @@ private:
 
 class ForwardTest : public ::testing::Test {
 protected:
+    isc::test::Sandbox sandbox;
 
     /// @brief Returns socket path (using either hardcoded path or env variable)
     /// @return path to the unix socket
     std::string getSocketPath() {
-
         std::string socket_path;
         const char* env = getenv("KEA_SOCKET_TEST_DIR");
         if (env) {
             socket_path = string(env) + "/test.unix";
         } else {
-            socket_path = string(TEST_DATA_BUILDDIR) + "/test.unix";
+            socket_path = sandbox.join("test.unix");
         }
         return (socket_path);
     }
@@ -532,18 +533,18 @@ ForwardTest::checkPushAndPop(int family, int type, int protocol,
     // reliable.  Instead we impose some reasonably large upper time limit of
     // blocking (normally it shouldn't even block at all; the limit is to
     // force the test to stop even if there's some bug and recv fails).
-    char recvbuf[sizeof(TEST_DATA)];
+    char recvbuf[sizeof(TEST_DATA) + 1] = { };
     sockaddr_storage ss;
     socklen_t sa_len = sizeof(ss);
     if (protocol == IPPROTO_UDP) {
-        EXPECT_EQ(sizeof(recvbuf),
-                  recvfrom(fwd_fd, recvbuf, sizeof(recvbuf),
+        EXPECT_EQ(sizeof(TEST_DATA),
+                  recvfrom(fwd_fd, recvbuf, sizeof(TEST_DATA),
                            setRecvDelay(fwd_fd), convertSockAddr(&ss),
                            &sa_len));
     } else {
         setNonBlock(client_sock.fd, false);
-        EXPECT_EQ(sizeof(recvbuf),
-                  recv(client_sock.fd, recvbuf, sizeof(recvbuf),
+        EXPECT_EQ(sizeof(TEST_DATA),
+                  recv(client_sock.fd, recvbuf, sizeof(TEST_DATA),
                        setRecvDelay(client_sock.fd)));
     }
     EXPECT_EQ(string(TEST_DATA), string(recvbuf));
