@@ -5,7 +5,7 @@
 if [ -e ${prefix}/share/kea/scripts/admin-utils.sh ]; then
     . ${prefix}/share/kea/scripts/admin-utils.sh
 else
-    . /home/wlodek/dev/kea_rel_1.4/src/bin/admin/admin-utils.sh
+    . /home/godfryd/isc/repos/kea-1.6.0-beta/src/bin/admin/admin-utils.sh
 fi
 
 VERSION=`pgsql_version "$@"`
